import sys
import matplotlib
matplotlib.use("Agg")
import numpy as np


import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages

# ceres modules
import os
_pkg_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
_data_dir = os.path.join(_pkg_dir, 'instruments', 'data')
_xc_masks_dir = os.path.join(_pkg_dir, 'utils', 'data', 'xc_masks') + '/'

baryc_dir = ''
ephemeris = 'DEc403'

from ceres3.instruments import uvesutils
from ceres3.utils import correlation
from ceres3.utils import globalutils as GLOBALutils
from ceres3.ext import Marsh

# other useful modules
import argparse
import ephem
from ceres3.utils import jplephem
from math import radians as rad
from astropy.io import fits as pyfits
import pickle
import scipy
import scipy.interpolate
from scipy import interpolate

import statsmodels.api as sm
lowess = sm.nonparametric.lowess


# Recive input parameters
parser = argparse.ArgumentParser()
parser.add_argument('directorio')
parser.add_argument('-avoid_plot', action="store_true", default=False)
parser.add_argument('-dirout',default='default')
parser.add_argument('-do_class', action="store_true", default=False)
parser.add_argument('-just_extract', action="store_true", default=False)
parser.add_argument('-npools', default=1)
parser.add_argument('-o2do',default='all')
parser.add_argument('-reffile',default='default')
parser.add_argument('-ofind', default='last')
parser.add_argument('-template', default='UVES_red_obs_exp')
parser.add_argument('-slit', default='0.7')

args = parser.parse_args()
dirin            = args.directorio
avoid_plot       = args.avoid_plot
dirout           = args.dirout
DoClass          = args.do_class
JustExtract      = args.just_extract
npools           = int(args.npools)
object2do        = args.o2do
reffile          = args.reffile
stst            = args.ofind
template        = args.template
slit            = args.slit

if dirin[-1] != '/':
    dirin = dirin + '/'

if dirout == 'default':
    dirout = dirin[:-1]+'_red/'

if not os.access(dirout,os.F_OK):
    os.makedirs(dirout, exist_ok=True)
if os.access(dirout+'proc',os.F_OK):
    os.system('rm -r '+dirout+'proc')
os.makedirs(dirout+'proc', exist_ok=True)

f_res = open(dirout+'proc/'+'results.txt','w')

if reffile == 'default':
    reffile = dirin+'reffile.txt'

####### GLOBAL VARIABLES #####
## perhaps put into options ##
force_pre_process  = False
force_flat_extract = False
force_flat_nor     = False
force_thar_extract = False
force_thar_wavcal  = False
force_tharxc       = False      
force_sci_extract  = False
force_spectral_file_build = True
force_stellar_pars = False
force_bac          = False
dumpargon          = False
minlines_glob      = 400

Inverse_m          = True
use_cheby          = True
MRMS               = 100   # max rms in m/s, global wav solution

trace_degree       = 6
Marsh_alg          = 0
ext_aperture       = 21
sky_aperture       = 31
NSigma_Marsh       = 10
NCosmic_Marsh      = 10
S_Marsh            = 0.4
N_Marsh            = 3      # grado polinomio

if template == 'UVES_red_obs_exp':
        min_extract_col      = 0
        min_extract_cols1    = [920,0,0,0,0,0,0,0,0,0,0,0,0,0,0]
        min_extract_cols2    = [1710,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]
        max_extract_col      = 4000
        max_extract_cols1    = [3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895]
        max_extract_cols2    = [3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3390]
elif template == 'UVES_dic2_obs_expfree':
        min_extract_col      = 0
        min_extract_cols1    = [920 ,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]
        min_extract_cols2    = [1710,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]
        max_extract_col      = 4000
        max_extract_cols1    = [3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895]
        max_extract_cols2    = [3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3895,3390]

xbinning           = 1
ybinning           = 1

ext_aperture = int(float(ext_aperture)/float(ybinning)) 

ncoef_x   = 5
ncoef_m   = 6
npar_wsol = (min(ncoef_x,ncoef_m) + 1) * (2*max(ncoef_x,ncoef_m) - min(ncoef_x,ncoef_m) + 2) // 2
ro0       = 90
order_gap = 2
order_dir  = os.path.join(_data_dir, 'uves_wavcals/')

#############################

print("\n\n\tUVES @VLT  PIPELINE\n")
print("\tRAW data is in ",dirin)
print("\tProducts of reduction will be in",dirout)
print('\n')

# file containing the log
log = dirout+'night.log'
biases, flats, orderref, ThAr_ref, sim_sci, ThAr_ref_dates, obnames, exptimes = uvesutils.FileClassify(dirin,log)

print('\tThis in the log of the night:\n')
f = open(log)
flines = f.readlines()
for line in flines:
        print('\t\t'+line[:-1])
print('\n')

if ((os.access(dirout+'Flat.fits',os.F_OK) == False) or \
    (os.access(dirout+'MasterBias.fits',os.F_OK) == False) or \
    (os.access(dirout+'trace.pkl',os.F_OK) == False)  or \
    (force_pre_process) ):
    print("\tNo previous pre-processing files or found")
    pre_process = 1
else:
    print("\tPre-processing files found, going straight to extraction")
    pre_process = 0

if (pre_process == 1):
    # median combine list of flats
    print("\t\tGenerating Master calibration frames...")

    MasterBias, RO_bias, GA_bias = uvesutils.MedianCombine(biases)
    hdu = pyfits.PrimaryHDU(MasterBias)
    if (os.access(dirout+'MasterBias.fits',os.F_OK)):
        os.remove(dirout+'MasterBias.fits')
    hdu.writeto(dirout+'MasterBias.fits')
    print("\t\t-> MasterBias: done!")

    Flat, RO_fl, GA_fl = uvesutils.MedianCombine(flats,bias=MasterBias)
    hdu = pyfits.PrimaryHDU(Flat)
    if (os.access(dirout+'Flat.fits',os.F_OK)):
        os.remove(dirout+'Flat.fits')
    hdu.writeto(dirout+'Flat.fits')
    print("\t\t-> MasterFlat: done!")

    hdu = pyfits.open(orderref[0])
    d1 = hdu[1].data - MasterBias[:,:,0]
    d2 = hdu[2].data - MasterBias[:,:,1]
    d1 = d1.T
    d2 = d2.T
    print("\tTracing echelle orders...")
    c_all1, nord1 = GLOBALutils.get_them(d1,int(0.2*ext_aperture),trace_degree,mode=1,nsigmas=10)
    print('\t\t'+str(nord1)+' orders for chip1 found...')
    c_all2, nord2 = GLOBALutils.get_them(d2,int(0.2*ext_aperture),trace_degree,mode=1,nsigmas=10, endat=2030)
    print('\t\t'+str(nord2)+' orders for chip2 found...')

    trace_dict = {'c_all1':c_all1,'nord1':nord1,\
                  'c_all2':c_all2,'nord2':nord2,\
                  'GA_bias': GA_bias, 'RO_bias': RO_bias,\
                  'GA_fl': GA_fl, 'RO_fl': RO_fl}
    pickle.dump(trace_dict, open(dirout+'trace.pkl', 'wb'))

else:
    trace_dict = pickle.load(open(dirout+'trace.pkl', 'rb'), encoding='latin1')
    c_all1 = trace_dict['c_all1']
    nord1  = trace_dict['nord1']
    c_all2 = trace_dict['c_all2']
    nord2  = trace_dict['nord2']

    GA_fl = trace_dict['GA_fl']
    RO_fl = trace_dict['RO_fl']
    GA_bias = trace_dict['GA_bias']
    RO_bias = trace_dict['RO_bias']

    Flat = pyfits.getdata(dirout+'Flat.fits')
    MasterBias = pyfits.getdata(dirout+'MasterBias.fits')

print('\n\tNormalization of Flat calibration frame:')
# Normalization of flat frame
if (os.access(dirout+'NFlat.fits',os.F_OK)==False) or (os.access(dirout+'Flat_spec.pkl',os.F_OK)==False) or force_flat_nor:

    ccdflat1 = uvesutils.ccd_flat(Flat[:,:,0],c_all1,sky_aperture)
    ccdflat2 = uvesutils.ccd_flat(Flat[:,:,1],c_all2,sky_aperture)
    ccdflat  = np.dstack((ccdflat1,ccdflat2))

    nflat1,sflat1 = uvesutils.normalize_flat(Flat[:,:,0],c_all1,sky_aperture)
    nflat2,sflat2 = uvesutils.normalize_flat(Flat[:,:,1],c_all2,sky_aperture)

    nflat = np.dstack((nflat1,nflat2))
    sflat = {'chip1':sflat1, 'chip2':sflat2}

    hdu = pyfits.PrimaryHDU(nflat)
    if (os.access(dirout+'NFlat.fits',os.F_OK)):
        os.remove(dirout+'NFlat.fits')
    hdu.writeto(dirout+'NFlat.fits')

    pickle.dump(sflat, open(dirout+'Flat_spec.pkl', 'wb'))
        
else:
    nflat = pyfits.getdata(dirout+'NFlat.fits')
    sflat = pickle.load(open(dirout+'Flat_spec.pkl', 'rb'), encoding='latin1')

print('\n\tExtraction of ThAr calibration frames:')
for fsim in ThAr_ref:
    print(fsim)
    thar_fits = dirout + fsim.split('/')[-1][:-4]+'spec.pkl'
    if ( os.access(thar_fits,os.F_OK) == False ) or ( force_thar_extract ):
        hthar = pyfits.open( fsim )
        dthar = (np.dstack((hthar[1].data,hthar[2].data)) -  MasterBias) / nflat
        print("\t\tNo previous extraction or extraction forced for ThAr file", fsim, "extracting...")
        dthar1 = dthar[:,:,0].T
        dthar2 = dthar[:,:,1].T 

        thar1_S = GLOBALutils.simple_extraction(dthar1,c_all1,ext_aperture,min_extract_cols1,max_extract_cols1,npools)
        thar2_S = GLOBALutils.simple_extraction(dthar2,c_all2,ext_aperture,min_extract_cols2,max_extract_cols2,npools)
        thar1_S = GLOBALutils.invert(thar1_S)
        thar2_S = GLOBALutils.invert(thar2_S)
        thar_S = {'chip1':thar1_S, 'chip2':thar2_S}
        pickle.dump(thar_S, open(thar_fits, 'wb'))

    else:
        print("\t\tThAr file", fsim, "all ready extracted, loading...")

print("\n\tWavelength solution of ThAr calibration spectra:")
for fsim in ThAr_ref:

    hthar = pyfits.open( fsim )
    thar_fits  = dirout + fsim.split('/')[-1][:-4]+'spec.pkl'
    wavsol_pkl = dirout + fsim.split('/')[-1][:-4]+'spec.wavsolpars.pkl'

    if ( os.access(wavsol_pkl,os.F_OK) == False ) or ( force_thar_wavcal ):

        print("\t\tWorking on initial ThAr file", fsim)
        mjd, mjd0 = uvesutils.mjd_fromheader2( hthar[0].header )
        thar_hd   = hthar[0].header
        thar_S    = pickle.load(open(thar_fits, 'rb'), encoding='latin1')

        lines_thar2       = thar_S['chip1']
        All_Pixel_Centers = np.array([])
        All_Wavelengths   = np.array([])
        All_Orders        = np.array([])
        All_Centroids     = np.array([])
        All_Sigmas        = np.array([])
        All_Intensities   = np.array([])

        for order in range(len(c_all1)):
            order_s = str(order)
            if (order < 10):
                order_s = '0'+str(order)
            thar_order_orig = lines_thar2[order]
            I = np.where(np.isnan(thar_order_orig))
            J = np.where(np.isnan(thar_order_orig)==False)
            thar_order_orig[I] = np.median(thar_order_orig[J])
            thar_order         = thar_order_orig - scipy.signal.medfilt(thar_order_orig,101)
            wei = np.ones(len(thar_order))
            if order == 0:
                do_xc = False
            else:
                do_xc = True
            coeffs_pix2wav, coeffs_pix2sigma, pixel_centers, wavelengths,\
            rms_ms, residuals, centroids, sigmas, intensities =\
                     GLOBALutils.Initial_Wav_Calibration(order_dir+'nborder_'+\
                     order_s+'.iwdat', thar_order, order, wei, rmsmax=100, \
                     minlines=20,FixEnds=False,Dump_Argon=dumpargon,sigmai=2.,\
                     Dump_AllLines=True, Cheby=use_cheby,porder=ncoef_x,do_xc=False, line_width=6,pixelization=True)

            if order == 7: 
                Global_ZP = GLOBALutils.Cheby_eval( coeffs_pix2wav, 0.5*(len(thar_order)-1), len(thar_order))

            All_Pixel_Centers = np.append( All_Pixel_Centers, pixel_centers )
            All_Wavelengths   = np.append( All_Wavelengths, wavelengths )
            All_Orders        = np.append( All_Orders, np.zeros( len(pixel_centers) ) + order )
            All_Centroids     = np.append( All_Centroids, centroids)
            All_Sigmas        = np.append( All_Sigmas, sigmas)
            All_Intensities   = np.append( All_Intensities, intensities )

        p0 = np.zeros( npar_wsol )
        p0[0] =  (7+ro0) * Global_ZP 
        p1, G_pix, G_ord, G_wav, II, rms_ms, G_res = \
            GLOBALutils.Fit_Global_Wav_Solution(All_Pixel_Centers, All_Wavelengths, All_Orders,\
                                                np.ones(All_Intensities.shape), p0, Cheby=use_cheby,\
                                                maxrms=MRMS, Inv=Inverse_m,minlines=200,order0=ro0, \
                                                ntotal=nord1,npix=len(thar_order),nx=ncoef_x,nm=ncoef_m)
        """
        equis = np.arange(len(thar_order))
        order = 0
        while order < nord1:
            m   = order + ro0
            chebs = GLOBALutils.Calculate_chebs(equis, m, order0=ro0,ntotal=nord1, \
                npix=len(equis), Inverse=Inverse_m,nx=ncoef_x,nm=ncoef_m)

            WavSol = (1.0/float(m)) * GLOBALutils.Joint_Polynomial_Cheby(p1,chebs,ncoef_x,ncoef_m)
            plot(WavSol,lines_thar2[order])
            order+=1
        show()
        """
        lines_thar2       = thar_S['chip2']
        All_Pixel_Centers2 = np.array([])
        All_Wavelengths2   = np.array([])
        All_Orders2        = np.array([])
        All_Centroids2     = np.array([])
        All_Sigmas2        = np.array([])
        All_Intensities2   = np.array([])

        for order in range(len(c_all2)):
            order_s = str(order)
            if (order < 10):
                order_s = '0'+str(order)

            thar_order_orig = lines_thar2[order]
            I = np.where(np.isnan(thar_order_orig))
            J = np.where(np.isnan(thar_order_orig)==False)
            thar_order_orig[I] = np.median(thar_order_orig[J])
            thar_order         = thar_order_orig - scipy.signal.medfilt(thar_order_orig,101)
            wei = np.ones(len(thar_order))

            coeffs_pix2wav, coeffs_pix2sigma, pixel_centers, wavelengths,\
            rms_ms, residuals, centroids, sigmas, intensities =\
                                         GLOBALutils.Initial_Wav_Calibration(order_dir+'nrorder_'+\
                                         order_s+'.iwdat', thar_order, order, wei, rmsmax=100, \
                                         minlines=20,FixEnds=False,Dump_Argon=dumpargon,sigmai=2.,\
                                         Dump_AllLines=True, Cheby=use_cheby,porder=ncoef_x,do_xc=False, line_width=6,pixelization=True)

            if order == 11: 
                Global_ZP = GLOBALutils.Cheby_eval( coeffs_pix2wav, 0.5*(len(thar_order)-1), len(thar_order))

            All_Pixel_Centers2 = np.append( All_Pixel_Centers2, pixel_centers )
            All_Wavelengths2   = np.append( All_Wavelengths2, wavelengths )
            All_Orders2        = np.append( All_Orders2, np.zeros( len(pixel_centers) ) + order )
            All_Centroids2     = np.append( All_Centroids2, centroids)
            All_Sigmas2        = np.append( All_Sigmas2, sigmas)
            All_Intensities2   = np.append( All_Intensities2, intensities )
        
                        #isz  = pixel_centers - sigmas
                        #der  = pixel_centers + sigmas
                        #isz  = GLOBALutils.Cheby_eval( coeffs_pix2wav, isz, len(thar_order))
                        #der  = GLOBALutils.Cheby_eval( coeffs_pix2wav, der, len(thar_order))
                        #sig  = 0.5*(der-isz)
                        #fwhm = 2.35 * sig
                        #resol = wavelengths / fwhm
                        #print order, np.median(resol)
                        #plot(pixel_centers,resol,'o')
                #show()
        #print dfgh

        p0 = np.zeros( npar_wsol )
        p0[0] =  (11 + ro0 + len(c_all1) + order_gap) * Global_ZP 
        p2, G_pix2, G_ord2, G_wav2, II2, rms_ms2, G_res2 = \
                        GLOBALutils.Fit_Global_Wav_Solution(All_Pixel_Centers2, All_Wavelengths2, All_Orders2,\
                                                np.ones(All_Intensities2.shape), p0, Cheby=use_cheby,\
                                                maxrms=MRMS, Inv=Inverse_m,minlines=400,order0=ro0 + len(c_all1) + order_gap, \
                                                ntotal=nord2,npix=len(thar_order),nx=ncoef_x,nm=ncoef_m)
        """
        equis = np.arange(len(thar_order))
        order = 0
        while order < nord2:
            m   = order + nord1 + order_gap + ro0
            chebs = GLOBALutils.Calculate_chebs(equis, m, order0=nord1 + order_gap + ro0,ntotal=nord2, \
                npix=len(equis), Inverse=Inverse_m,nx=ncoef_x,nm=ncoef_m)
            WavSol = (1.0/float(m)) * GLOBALutils.Joint_Polynomial_Cheby(p2,chebs,ncoef_x,ncoef_m)
            plot(WavSol,lines_thar2[order])
            order+=1
        show()
        """
        #for i in np.unique(G_ord2):
        #    I = np.where(G_ord2 == i)[0]
        #    plot(G_wav2[I],G_res2[I],'.')
        #show()
        pdict = {'p1':p1,'mjd':mjd, 'G_pix':G_pix, 'G_ord':G_ord, 'G_wav':G_wav, 'II':II, 'rms_ms':rms_ms,\
                 'G_res':G_res, 'All_Centroids':All_Centroids, 'All_Orders':All_Orders, 'All_Sigmas':All_Sigmas,\
                 'p2':p2, 'G_pix2':G_pix2, 'G_ord2':G_ord2, 'G_wav2':G_wav2, 'II2':II2, 'rms_ms2':rms_ms2,\
                 'G_res2':G_res2, 'All_Centroids2':All_Centroids2, 'All_Orders2':All_Orders2, 'All_Sigmas2':All_Sigmas2}
        pickle.dump(pdict, open(wavsol_pkl, 'wb'))

    else:
        print("\t\tUsing previously computed wavelength solution in file",wavsol_pkl                    )

print('\n\tThe following targets will be processed:')
new_list = []
if object2do == 'all':
        new_list = sim_sci
else:
        for fsim in sim_sci:
                hd = pyfits.getheader(fsim)
                if object2do in hd['OBJECT']:
                        new_list.append(fsim)

for fsim in new_list:
        hd = pyfits.getheader(fsim)
        print('\t\t'+str(hd['OBJECT']))

# Does any image have a special requirement for dealing with the moonlight?
if os.access(dirin + 'moon_corr.txt', os.F_OK):
    fmoon = open(dirin + 'moon_corr.txt','r')
    moon_lns = fmoon.readlines()
    spec_moon = []
    use_moon = []
    for line in moon_lns:
        spec_moon.append(line.split()[0])
        if line.split()[1] == '0':
            use_moon.append(False)
        else:
            use_moon.append(True)
else:
    spec_moon = []
    use_moon = []

spec_moon = np.array(spec_moon)
use_moon  = np.array(use_moon)

# Extract all Science files
for fsim in new_list:
    print('\n')
    print("\t--> Working on image: ", fsim)
    know_moon = False

    if fsim.split('/')[-1] in spec_moon:
        I = np.where(fsim.split('/')[-1] == spec_moon)[0]
        know_moon = True
        here_moon = use_moon[I]

    h = pyfits.open(fsim)
    mjd,mjd0      = uvesutils.mjd_fromheader2(h[0].header)    
    ron1,gain1 = float(h[1].header['HIERARCH ESO DET OUT1 RON']),float(h[1].header['HIERARCH ESO DET OUT1 GAIN'])
    ron2,gain2 = float(h[2].header['HIERARCH ESO DET OUT1 RON']),float(h[2].header['HIERARCH ESO DET OUT1 GAIN'])

    # Object name
    obname    = h[0].header['OBJECT'].replace(' ','')

    print("\t\tObject name:",obname)
    # Open file, trim, overscan subtract and MasterBias subtract
    bac_fits = dirout + 'BAC_' + fsim.split('/')[-1][:-4]+'.fits'
    sky_fits = dirout + 'SKY_' + fsim.split('/')[-1][:-4]+'.fits'
    #data = hiresutils.OverscanTrim( h[chip].data, h[chip].header['DATASEC'] ).T
    #c_new,pshift = GLOBALutils.retrace( data, c_all, span=15 )
    #print '\t\ty shift = ', pshift, 'pxs'

    data = np.dstack((h[1].data,h[2].data)) -  MasterBias
    centers1 = np.zeros((len(c_all1),data.shape[0]))
    centers2 = np.zeros((len(c_all2),data.shape[0]))

    for i in range(len(c_all1)):
        centers1[i] = np.polyval(c_all1[i],np.arange(data.shape[0]))
    for i in range(len(c_all2)):
        centers2[i] = np.polyval(c_all2[i],np.arange(data.shape[0]))

    print('\t\tEstimating background of scattered light...')
    force_bac=False
    if (os.access(bac_fits,os.F_OK) == False) or force_bac:
        bac1 = GLOBALutils.get_scat(data[:,:,0].T,centers1,span=sky_aperture)
        bac2 = GLOBALutils.get_scat(data[:,:,1].T,centers2,span=sky_aperture)
        bac = np.dstack((bac1.T,bac2.T))
        bach = pyfits.PrimaryHDU(bac)
        if os.access(bac_fits,os.F_OK):
            os.system('rm ' + bac_fits)
        bach.writeto(bac_fits)
    else:
        bac = pyfits.getdata(bac_fits)

    data = data - bac

    if (os.access(sky_fits,os.F_OK) == False) or force_bac:
        sky1 = uvesutils.get_sky(data[:,:,0].T,centers1,ext_aperture,sky_aperture)
        sky2 = uvesutils.get_sky(data[:,:,1].T,centers2,ext_aperture,sky_aperture)
        sky = np.dstack((sky1.T,sky2.T))

        bach = pyfits.PrimaryHDU(sky)
        if os.access(sky_fits,os.F_OK):
            os.system('rm ' + sky_fits)
        bach.writeto(sky_fits)
    else:
        sky = pyfits.getdata(sky_fits)

    data = data - sky 
    data = data / nflat

    ra  = float(h[0].header['RA'])
    dec = float(h[0].header['DEC'])

    ra2,dec2 = GLOBALutils.getcoords(obname,mjd,filen=reffile)
    if ra2 !=0 and dec2 != 0:
        ra = ra2
        dec = dec2
    else:
        print('\t\tUsing the coordinates found in the image header.')

    # set observatory parameters
    altitude    = float(h[0].header['ESO TEL GEOELEV'])
    latitude    = float(h[0].header['ESO TEL GEOLAT'])
    longitude   = float(h[0].header['ESO TEL GEOLON'])
    epoch       = 2000.

    iers                    = GLOBALutils.JPLiers( baryc_dir, mjd-999.0, mjd+999.0 )
    obsradius, R0           = GLOBALutils.JPLR0( latitude, altitude)
    obpos                   = GLOBALutils.obspos( longitude, obsradius, R0 )

    jplephem.set_ephemeris_dir( baryc_dir , ephemeris )
    jplephem.set_observer_coordinates( float(obpos[0]), float(obpos[1]), float(obpos[2]) )

    res         = jplephem.doppler_fraction(float(ra/15.0), float(dec), int(mjd), float(mjd%1), 1, 0.0)
    lbary_ltopo = 1.0 + res['frac'][0]
    bcvel_baryc = ( lbary_ltopo - 1.0 ) * 2.99792458E5

    print("\t\tBarycentric velocity:", bcvel_baryc)

    res  = jplephem.pulse_delay(ra/15.0, dec, int(mjd), mjd%1, 1, 0.0)
    mbjd = mjd + res['delay'][0] / (3600.0 * 24.0)

    # Moon Phase Calculations
    gobs      = ephem.Observer()  
    gobs.name = 'VLT'
    gobs.lat  = rad(latitude)  # lat/long in decimal degrees  
    gobs.long = rad(longitude) 
    gobs.date = h[0].header['DATE-OBS'].replace('T',' ')
    mephem    = ephem.Moon()
    mephem.compute(gobs)
    Mcoo        = jplephem.object_track("Moon", int(mjd), float(mjd%1), 1, 0.0)
    Mp   = jplephem.barycentric_object_track("Moon", int(mjd), float(mjd%1), 1, 0.0)
    Sp   = jplephem.barycentric_object_track("Sun", int(mjd), float(mjd%1), 1, 0.0)
    res  = jplephem.object_doppler("Moon", int(mjd), mjd%1, 1, 0.0)
    lunation,moon_state,moonsep,moonvel = GLOBALutils.get_lunar_props(ephem,gobs,Mcoo,Mp,Sp,res,ra,dec)
    refvel = bcvel_baryc + moonvel
    print('\t\tRadial Velocity of sacttered moonlight:',refvel)

    sci_fits        = dirout + fsim.split('/')[-1][:-4]+'spec.pkl'
    sci_fits_simple = dirout + fsim.split('/')[-1][:-4]+'spec.simple.pkl'
    P_fits          = dirout + 'P_' + fsim.split('/')[-1][:-4]+'spec.fits'

    if ( os.access(sci_fits,os.F_OK) == False ) or ( os.access(sci_fits_simple,os.F_OK) == False ) or \
       ( force_sci_extract ):
        print("\t\tNo previous extraction or extraction forced for science file", fsim, "extracting...")
        #"""
        print("\t\t\tweights for chip1...")
        P1 = GLOBALutils.obtain_P(data[:,:,0].T,c_all1,ext_aperture,ron1,gain1,NSigma_Marsh, S_Marsh, \
                                    N_Marsh, Marsh_alg, min_extract_cols1,max_extract_cols1, npools)
        print("\t\t\tweights for chip2...")
        P2 = GLOBALutils.obtain_P(data[:,:,1].T,c_all2,ext_aperture,ron2,gain2,NSigma_Marsh, S_Marsh, \
                                    N_Marsh, Marsh_alg, min_extract_cols2,max_extract_cols2, npools)
        P = np.dstack((P1,P2))
        #"""
        #P = pyfits.getdata(P_fits)
        print("\t\t\tsimple extraction for chip1...")
        sci_Ss1 = GLOBALutils.simple_extraction(data[:,:,0].T,c_all1,ext_aperture,min_extract_cols1,max_extract_cols1,npools)
        print("\t\t\tsimple extraction for chip2...")
        sci_Ss2 = GLOBALutils.simple_extraction(data[:,:,1].T,c_all2,ext_aperture,min_extract_cols2,max_extract_cols2,npools)
        
        sci_Ss1 = GLOBALutils.invert(sci_Ss1)
        sci_Ss2 = GLOBALutils.invert(sci_Ss2)
        #"""
        print("\t\t\t\\optimal extraction for chip1...")
        sci_S1  = GLOBALutils.optimal_extraction(data[:,:,0].T,P1,c_all1,ext_aperture,ron1,gain1,S_Marsh,NCosmic_Marsh,\
                 min_extract_cols1,max_extract_cols1,npools)
        print("\t\t\t\\optimal extraction for chip2...")
        sci_S2  = GLOBALutils.optimal_extraction(data[:,:,1].T,P2,c_all2,ext_aperture,ron2,gain2,S_Marsh,NCosmic_Marsh,\
         min_extract_cols2,max_extract_cols2,npools)
        #"""
        #sci_S = pickle.load(open(sci_fits, 'rb'), encoding='latin1')
        #sci_S1 = sci_S['chip1']
        #sci_S2 = sci_S['chip2']

        sci_S1 = GLOBALutils.invert(sci_S1)
        sci_S2 = GLOBALutils.invert(sci_S2)
        
        if (os.access(P_fits,os.F_OK)):
            os.remove( P_fits )
        hdu = pyfits.PrimaryHDU( P )
        hdu.writeto( P_fits )
        sci_S  = {'chip1':sci_S1, 'chip2':sci_S2}
        pickle.dump(sci_S, open(sci_fits, 'wb'))
        
        sci_Ss = {'chip1':sci_Ss1, 'chip2':sci_Ss2}
        pickle.dump(sci_Ss, open(sci_fits_simple, 'wb'))
        
    else:
        print('\t\t'+fsim+" has already been extracted, reading in product fits files...")
        sci_S  = pickle.load(open(sci_fits, 'rb'), encoding='latin1')
        sci_Ss = pickle.load(open(sci_fits_simple, 'rb'), encoding='latin1')

    fout = 'proc/'+ obname + '_' + h[0].header['DATE-OBS'] + '_' + 'sp.fits'
    if ( os.access(dirout+fout ,os.F_OK) == False ) or (force_spectral_file_build):

        spec = np.zeros((11, nord1 + nord2, data.shape[0]))
        hdu = pyfits.PrimaryHDU( spec )
        hdu = GLOBALutils.update_header(hdu,'HIERARCH MJD', mjd)
        hdu = GLOBALutils.update_header(hdu,'HIERARCH MBJD', mbjd)
        hdu = GLOBALutils.update_header(hdu,'HIERARCH SHUTTER START DATE', h[0].header['DATE-OBS'].split('T')[0] )
        hdu = GLOBALutils.update_header(hdu,'HIERARCH SHUTTER START UT',  h[0].header['DATE-OBS'].split('T')[1])
        hdu = GLOBALutils.update_header(hdu,'HIERARCH TEXP (S)',h[0].header['EXPTIME'])
        hdu = GLOBALutils.update_header(hdu,'HIERARCH BARYCENTRIC CORRECTION (KM/S)', bcvel_baryc)
        hdu = GLOBALutils.update_header(hdu,'HIERARCH (LAMBDA_BARY / LAMBDA_TOPO)', lbary_ltopo)    
        hdu = GLOBALutils.update_header(hdu,'HIERARCH TARGET NAME', obname)
        hdu = GLOBALutils.update_header(hdu,'HIERARCH RA',h[0].header['RA'])
        hdu = GLOBALutils.update_header(hdu,'HIERARCH DEC',h[0].header['DEC'])
        hdu = GLOBALutils.update_header(hdu,'HIERARCH RA BARY',ra)
        hdu = GLOBALutils.update_header(hdu,'HIERARCH DEC BARY',dec)
        hdu = GLOBALutils.update_header(hdu,'HIERARCH EQUINOX',2000.)
        hdu = GLOBALutils.update_header(hdu,'HIERARCH OBS LATITUDE',latitude)
        hdu = GLOBALutils.update_header(hdu,'HIERARCH OBS LONGITUDE',longitude)
        hdu = GLOBALutils.update_header(hdu,'HIERARCH OBS ALTITUDE',altitude)
        hdu = GLOBALutils.update_header(hdu,'HIERARCH TARG AIRMASS',h[0].header['ESO OBS AIRM'])

        models_path = os.path.join(_pkg_dir, 'data', 'COELHO_MODELS', 'R_40000b/')
        ref_RES = 60000.
        # get ThAr closest in time
        im = np.argmin(np.absolute(np.array(ThAr_ref_dates) - mjd))
        pkl_wsol = dirout + ThAr_ref[im].split('/')[-1][:-4]+'spec.wavsolpars.pkl'
        print("\t\tUnpickling wavelength solution from", pkl_wsol, " ...")
        wsol_dict = pickle.load(open(pkl_wsol, 'rb'), encoding='latin1')

        # Apply new wavelength solution including barycentric correction
        equis = np.arange( data.shape[0] )        
        for order in range(nord1):
            #order = 16
            #print order
            m = order + ro0
            chebs = GLOBALutils.Calculate_chebs(equis, m, npix=data.shape[0],\
                    order0=ro0, ntotal=nord1, Inverse=Inverse_m,nx=ncoef_x,nm=ncoef_m)
            WavSol = lbary_ltopo * (1.0/m) * GLOBALutils.Joint_Polynomial_Cheby(wsol_dict['p1'],chebs,ncoef_x,ncoef_m)   
            spec[0,order,:] = GLOBALutils.ToVacuum(WavSol)
            spec[1,order,:] = sci_S['chip1'][order,1,:]
            #spec[1,order,:300] = 0.
            spec[2,order,:] = sci_S['chip1'][order,2,:]
            spec[3,order,:] = spec[1,order,:] / sflat['chip1'][order][::-1]
            spec[4,order,:] = spec[2,order,:] * sflat['chip1'][order][::-1] ** 2
            ccoef = GLOBALutils.get_cont_single(spec[0,order],spec[3,order],spec[4,order],ll=1.5,lu=5,nc=3)
            ratio = np.polyval(ccoef, spec[0,order])
            L     = np.where( spec[1,order,:] != 0 )
            spec[5,order,:][L] = spec[3,order,:][L] / ratio[L]
            spec[3,order,:][L] = spec[3,order,:][L] / ratio[L]
            spec[6,order,:][L] = spec[4,order,:][L] * (ratio[L] ** 2 )
            spec[7,order,:][L] = ratio[L]
            spec[8,order,:][L] = ratio[L] * sflat['chip1'][order][::-1][L] / np.sqrt( ratio[L] * sflat['chip1'][order][::-1][L] / gain1 + (ron1/gain1)**2 )
            rI = np.where(spec[5,order] > 1. + 5./spec[8,order])
            spec[5,order,rI] = 1.
            lI = np.where(spec[5,order] < -5./spec[8,order])
            spec[5,order,lI] = 0.

            spl           = scipy.interpolate.splrep(np.arange(len(WavSol)), WavSol,k=3)
            dlambda_dx    = scipy.interpolate.splev(np.arange(WavSol.shape[0]), spl, der=1)
            NN            = np.average(dlambda_dx)
            dlambda_dx    /= NN
            # clean-up of CRs in continuum-normalized spectrum. Set troublesome pixels to 1
            medflx = np.zeros(len(spec[5,order]))
            nI = np.where(np.isnan(spec[5,order])==False)[0]
            nJ = np.where(np.isnan(spec[5,order])==True)[0]
            spec[6,order,:][L] = spec[8,order,:][L] ** 2
            spec[9,order,:][L] = spec[5,order,:][L] * (dlambda_dx[L] ** 1) 
            spec[10,order,:][L] = spec[6,order,:][L] / (dlambda_dx[L] ** 2)


        for order in range(nord2):
            order2 = order + nord1
            m = order + ro0 + nord1 + order_gap
            chebs = GLOBALutils.Calculate_chebs(equis, m, npix=data.shape[0],\
            order0=ro0 + nord1 + order_gap, ntotal=nord2, Inverse=Inverse_m,nx=ncoef_x,nm=ncoef_m)
            WavSol = lbary_ltopo * (1.0/m) * GLOBALutils.Joint_Polynomial_Cheby(wsol_dict['p2'],chebs,ncoef_x,ncoef_m)   
            spec[0,order2,:] = GLOBALutils.ToVacuum(WavSol)
            spec[1,order2,:] = sci_S['chip2'][order,1,:]
            #spec[1,order2,:300] = 0.
            spec[2,order2,:] = sci_S['chip2'][order,2,:]
            spec[3,order2,:] = spec[1,order2,:] / sflat['chip2'][order][::-1]
            spec[4,order2,:] = spec[2,order2,:] * sflat['chip2'][order][::-1] ** 2
            ccoef = GLOBALutils.get_cont_single(spec[0,order2],spec[3,order2],spec[4,order2],ll=1.5,lu=5,nc=3)
            ratio = np.polyval(ccoef, spec[0,order2])
            L     = np.where( spec[1,order2,:] != 0 )
            spec[5,order2,:][L] = spec[3,order2,:][L] / ratio[L]
            spec[3,order2,:][L] = spec[3,order2,:][L] / ratio[L]
            spec[6,order2,:][L] = spec[4,order2,:][L] * (ratio[L] ** 2 )
            spec[7,order2,:][L] = ratio[L]
            spec[8,order2,:][L] = ratio[L] * sflat['chip2'][order][::-1][L] / np.sqrt( ratio[L] * sflat['chip2'][order][::-1][L] / gain2 + (ron2/gain2)**2 )
            rI = np.where(spec[5,order2] > 1. + 5./spec[8,order2])
            #lI = np.where(spec[5,order2] < -8./spec[8,order2])
            lI = np.where(spec[5,order2] < -5./spec[8,order2])
            spec[5,order2,rI] = 1.
            spec[5,order2,lI] = 0.

            spl           = scipy.interpolate.splrep(np.arange(len(WavSol)), WavSol,k=3)
            dlambda_dx    = scipy.interpolate.splev(np.arange(WavSol.shape[0]), spl, der=1)
            NN            = np.average(dlambda_dx)
            dlambda_dx    /= NN
            # clean-up of CRs in continuum-normalized spectrum. Set troublesome pixels to 1
            medflx = np.zeros(len(spec[5,order2]))
            nI = np.where(np.isnan(spec[5,order2])==False)[0]
            nJ = np.where(np.isnan(spec[5,order2])==True)[0]
            spec[6,order2,:][L] = spec[8,order2,:][L] ** 2
            spec[9,order2,:][L] = spec[5,order2,:][L] * (dlambda_dx[L] ** 1) 
            spec[10,order2,:][L] = spec[6,order2,:][L] / (dlambda_dx[L] ** 2)

    if (not JustExtract):
        if DoClass:
            # spectral analysis
            print("\t\tSpectral Analysis...")
            query_success = False
            # First, query SIMBAD with the object name
            query_success,sp_type_query = GLOBALutils.simbad_query_obname(obname)
            # Now, query SIMBAD by coordinates if above not successful
            if (not query_success):
                query_success,sp_type_query = GLOBALutils.simbad_query_coords('12:00:00','00:00:00')
            print("\t\t\tSpectral type returned by SIMBAD query:",sp_type_query)

            hdu = GLOBALutils.update_header(hdu,'HIERARCH SIMBAD SPTYP', sp_type_query)

            pars_file = dirout + fsim.split('/')[-1][:-4]+'_stellar_pars.txt'

            if os.access(pars_file,os.F_OK) == False or force_stellar_pars:
                print("\t\t\tEstimating atmospheric parameters:")
                Rx = np.around(1./np.sqrt(1./40000.**2 - 1./ref_RES**2))
                spec2 = spec.copy()
                for i in range(spec.shape[1]):
                    IJ = np.where(spec[5,i]!=0.)[0]
                    spec2[5,i,IJ] = GLOBALutils.convolve(spec[0,i,IJ],spec[5,i,IJ],Rx)
                T_eff, logg, Z, vsini, vel0, ccf = correlation.CCF(spec2,model_path=models_path,npools=npools)
                line = "%6d %4.1f %4.1f %8.1f %8.1f\n" % (T_eff,logg, Z, vsini, vel0)
                f = open(pars_file,'w')
                f.write(line)
                f.close()
                       
            else:
                print("\t\t\tAtmospheric parameters loaded from file:")
                T_eff, logg, Z, vsini, vel0 = np.loadtxt(pars_file,unpack=True)

            print("\t\t\t\tT_eff=",T_eff,"log(g)=",logg,"Z=",Z,"vsin(i)=",vsini,"vel0",vel0)

        else:
            T_eff, logg, Z, vsini, vel0 = -999,-999,-999,-999,-999

        T_eff_epoch = T_eff
        logg_epoch  = logg
        Z_epoch     = Z
        vsini_epoch = vsini
        vel0_epoch  = vel0
        hdu = GLOBALutils.update_header(hdu,'HIERARCH TEFF', float(T_eff))
        hdu = GLOBALutils.update_header(hdu,'HIERARCH LOGG', float(logg))
        hdu = GLOBALutils.update_header(hdu,'HIERARCH Z', Z)
        hdu = GLOBALutils.update_header(hdu,'HIERARCH VSINI', vsini)
        hdu = GLOBALutils.update_header(hdu,'HIERARCH VEL0', vel0)

        print("\t\tRadial Velocity analysis:")
        # assign mask
        sp_type, mask = GLOBALutils.get_mask_reffile(obname,reffile=reffile,base=_xc_masks_dir)
        print("\t\t\tWill use",sp_type,"mask for CCF.")
                
        velw  = 300
        velsh = 3.
            
        # Read in mask
        ml, mh, weight = np.loadtxt(mask,unpack=True)
        ml_v = GLOBALutils.ToVacuum( ml )
        mh_v = GLOBALutils.ToVacuum( mh )
       
        # make mask larger accounting for factor ~2.5 lower res in FEROS w/r to HARPS
        av_m = 0.5*( ml_v + mh_v )
        ml_v -= 1.5*(av_m - ml_v)
        mh_v += 1.5*(mh_v - av_m)
        mask_hw_kms = (GLOBALutils.Constants.c/1e3) * 0.5*(mh_v - ml_v) / av_m

        disp = GLOBALutils.get_disp(obname, reffile=reffile)
        if disp == 0:
            known_sigma = False
            if vsini != -999 and vsini != 0.:
                disp = vsini
            else:
                disp = 3.
        else:
            known_sigma = True

        mask_hw_wide = av_m * disp / (GLOBALutils.Constants.c/1.0e3)
        ml_v = av_m - mask_hw_wide
        mh_v = av_m + mask_hw_wide 

        print('\t\t\tComputing the CCF...')
        cond = True
        while (cond):
            # first rough correlation to find the minimum
            vels, xc_full, sn, nlines_ccf, W_ccf = \
                    GLOBALutils.XCor(spec, ml_v, mh_v, weight, 0, lbary_ltopo, vel_width=velw,vel_step=velsh,\
                                          spec_order=9,iv_order=10,sn_order=8,max_vel_rough=velw)

            xc_av = GLOBALutils.Average_CCF(xc_full, sn, sn_min=3., Simple=True, W=W_ccf)
            yy     = scipy.signal.medfilt(xc_av,11)
            pred = lowess(yy, vels,frac=0.4,it=10,return_sorted=False)
            tck1 = scipy.interpolate.splrep(vels,pred,k=1)
            xc_av_orig = xc_av.copy()
            xc_av /= pred
            vel0_xc = vels[ np.argmin( xc_av ) ]
            rvels, rxc_av, rpred, rxc_av_orig, rvel0_xc = vels.copy(), xc_av.copy(), pred.copy(), xc_av_orig.copy(), vel0_xc
            xc_av_rough = xc_av
            vels_rough  = vels
                
            vel_width = np.maximum( 20.0, 6*disp )

            vels, xc_full, sn, nlines_ccf, W_ccf =\
                    GLOBALutils.XCor(spec, ml_v, mh_v, weight, vel0_xc, lbary_ltopo, vel_width=vel_width,vel_step=0.2,\
                                          spec_order=9,iv_order=10,sn_order=8,max_vel_rough=velw)
                
            xc_av = GLOBALutils.Average_CCF(xc_full, sn, sn_min=3., Simple=True, W=W_ccf)
            pred = scipy.interpolate.splev(vels,tck1)
            xc_av /= pred
                
            if sp_type == 'M5':
                moon_sig = 2.5
            elif sp_type == 'K5':
                moon_sig = 3.3
            else:
                moon_sig = 4.5

            p1,XCmodel,p1gau,XCmodelgau,Ls2 = GLOBALutils.XC_Final_Fit( vels, xc_av , sigma_res = 4, horder=8, moonv = refvel, moons = moon_sig, moon = False)

            moonmatters = False
            if (know_moon and here_moon):
                moonmatters = True
                ismoon = True
                confused = False
                p1_m,XCmodel_m,p1gau_m,XCmodelgau_m,Ls2_m = GLOBALutils.XC_Final_Fit( vels, xc_av , sigma_res = 4, horder=8, moonv = refvel, moons = moon_sig, moon = True)
                moon_flag = 1
            else:
                confused = False
                ismoon = False
                p1_m,XCmodel_m,p1gau_m,XCmodelgau_m,Ls2_m = p1,XCmodel,p1gau,XCmodelgau,Ls2
                moon_flag = 0

            SP = GLOBALutils.calc_bss2(vels,xc_av,p1gau)
            #SP = bspan[0]

            #print p1gau[1]
            if (not known_sigma):
                disp = np.floor(p1gau[2])
                if (disp < 3.0): 
                    disp = 3.0
                mask_hw_wide = av_m * disp / (GLOBALutils.Constants.c/1.0e3)
                ml_v = av_m - mask_hw_wide
                mh_v = av_m + mask_hw_wide            
                known_sigma = True
            else:
                cond = False

            BSerr = -999.00

        xc_dict = {'vels':vels,'xc_av':xc_av,'XCmodelgau':XCmodelgau,'Ls2':Ls2,'refvel':refvel,\
                       'rvels':rvels,'rxc_av':rxc_av,'rpred':rpred,'rxc_av_orig':rxc_av_orig,\
                       'rvel0_xc':rvel0_xc,'xc_full':xc_full, 'p1':p1, 'sn':sn, 'p1gau':p1gau,\
                       'p1_m':p1_m,'XCmodel_m':XCmodel_m,'p1gau_m':p1gau_m,'Ls2_m':Ls2_m,\
                       'XCmodelgau_m':XCmodelgau_m}

        moon_dict = {'moonmatters':moonmatters,'moon_state':moon_state,'moonsep':moonsep,\
                         'lunation':lunation,'mephem':mephem,'texp':h[0].header['EXPTIME']}


        pkl_xc = dirout + fsim.split('/')[-1][:-4]+obname+'_XC_'+sp_type+'.pkl'
        pickle.dump(xc_dict, open(pkl_xc, 'wb'))

        ccf_pdf = dirout + 'proc/' + fsim.split('/')[-1][:-4] + obname + '_XCs_' + sp_type + '.pdf'

        if not avoid_plot:
            GLOBALutils.plot_CCF(xc_dict,moon_dict,path=ccf_pdf)


        airmass  = h[0].header['ESO OBS AIRM']
        seeing   = -999
        TEXP = h[0].header['EXPTIME']    
        SNR_5130 = np.median(spec[8,13,1900:2101] )

        if sp_type=='G2':
            if T_eff < 6000:
                D = 0.32815
                C = 0.00453
            else:
                D = 0.32815
                C = 0.00453
        elif  sp_type == 'K5':
            D = 0.27404
            C = 0.00433
        else:
            D = 0.27404
            C = 0.00433

        BSerr = D / float(np.round(SNR_5130)) + C
        RVerr = 0.5

        RV     = np.around(p1gau_m[1],4)  
        BS     = np.around(SP,4)   
        RVerr2 = np.around(RVerr,4)
        BSerr  = np.around(BSerr,4)
        print('\t\t\tRV = '+str(RV)+' +- '+str(RVerr2))
        print('\t\t\tBS = '+str(BS)+' +- '+str(BSerr))

        bjd_out = 2400000.5 + mbjd
        T_eff_err = 100
        logg_err = 0.5
        Z_err = 0.5
        vsini_err = 2
        XC_min = np.abs(np.around(np.min(XCmodel),2))

        SNR_5130 = np.around(SNR_5130)
        SNR_5130_R = np.around(SNR_5130*np.sqrt(2.3))
        # write to output
        disp_epoch = np.around(p1gau_m[2],1)
        hdu = GLOBALutils.update_header(hdu,'RV', RV)
        hdu = GLOBALutils.update_header(hdu,'RV_E', RVerr2)
        hdu = GLOBALutils.update_header(hdu,'BS', BS)
        hdu = GLOBALutils.update_header(hdu,'BS_E', BSerr)
        hdu = GLOBALutils.update_header(hdu,'DISP', disp_epoch)
        hdu = GLOBALutils.update_header(hdu,'SNR', SNR_5130)
        hdu = GLOBALutils.update_header(hdu,'SNR_R', SNR_5130_R)
        hdu = GLOBALutils.update_header(hdu,'INST', 'UVES')
        hdu = GLOBALutils.update_header(hdu,'RESOL', ref_RES)
        hdu = GLOBALutils.update_header(hdu,'PIPELINE', 'CERES')
        hdu = GLOBALutils.update_header(hdu,'XC_MIN', XC_min)
        hdu = GLOBALutils.update_header(hdu,'BJD_OUT', bjd_out)
        line_out = "%-15s %18.8f %9.4f %7.4f %9.3f %5.3f   uves   ceres   %8d %6d %5.2f %5.2f %5.1f %4.2f %5.2f %6.1f %4d %s\n"%\
                      (obname, bjd_out, RV, RVerr2, BS, BSerr, ref_RES, T_eff_epoch, logg_epoch, Z_epoch, vsini_epoch, XC_min, disp_epoch,\
                       TEXP, SNR_5130_R, ccf_pdf)
        f_res.write(line_out)
        if (os.access( dirout + fout,os.F_OK)):
            os.remove( dirout + fout)
        hdu.writeto( dirout + fout )

f_res.close()



import sys
import math
from astropy.io import fits as pyfits
import numpy
import scipy
from scipy import optimize
from ceres3.ext import FunNorm
import time
import os
import matplotlib.pyplot as plt
import gc
import warnings

def division(L,F,n):

    x = int(len(L)//n)
    vec = numpy.zeros((2,x),float)

    i = 0
    while i < x:

        if i < x-1:
            vec[0][i] = numpy.mean(L[i*n:(i+1)*n])
            vec[1][i] = numpy.median(F[i*n:(i+1)*n])
        elif i == x-1:
            vec[0][i] = numpy.mean(L[i*n:len(L)])
            vec[1][i] = numpy.median(F[i*n:len(F)])

        i = i+1

    g = [-0.1,1.0]
    if vec.shape[1] > 2:
        ajuste = optimize.leastsq(res_lin,g,args=(vec[1][:],vec[0][:]))

    else:
        ajuste = optimize.leastsq(res_lin,g,args=(F,L))

    return ajuste

def lin(params,x):
    m = params[0]
    n = params[1]
    y = m*x+n
    return y

def res_lin(params,y,x):
    return y-lin(params,x)

def curva(params,x):
    a_0=params[0]
    a_1=params[1]
    a_2=params[2]
    a_3=params[3]
    a_4=params[4]

    y=a_0+a_1*x+a_2*x*x+a_3*(x**3)+a_4*(x**4)
    return y

def res_cur(params,y,x):
    return y-curva(params,x)

def NORM(L,F,orden=2):

    ordenes = L.shape[0]
    pixeles = L.shape[1]

    guess = [-0.1,1.0]
    guess2 = [1,0.1,0.1,0.1,0.1]

    div = 4
    rec = 20

    para = numpy.zeros( (2,div),float )
    continuo = numpy.zeros( (2,ordenes),list )
    strl = numpy.zeros( (6,2),float )
    laro = numpy.zeros( ordenes,int )

    #strli,strlf = numpy.loadtxt('/home/rabrahm/Desktop/corr2/strong_lines.dat',dtype=None,usecols=(0,1),unpack=True)
    strli,strlf = numpy.loadtxt('../utils/Continuum/strong_lines.dat',dtype=None,usecols=(0,1),unpack=True)

    esto = numpy.zeros( (2,ordenes),list )
    ajent = numpy.zeros( (ordenes,pixeles),float )
    unos = numpy.zeros(pixeles,float)
    NORMALIZADO = numpy.zeros(ordenes,list)
    unos = unos+1.0

    Lfilter = numpy.zeros(ordenes,list)
    Ffilter = numpy.zeros(ordenes,list)
    coefs = numpy.zeros([ordenes,orden+1])

    o=0
    while o<ordenes:

        Lmien = numpy.zeros(pixeles,float)
        Fmien = numpy.zeros(pixeles,float)

        """
        elemino flujos nulos
        """

        I = numpy.where( F[o] != 0.0 )[0]
        Llimpio = L[o][I]
        Flimpio = F[o][I]

        """
        hago filtro de 5 pixeles
        """

        i = 0
        k = 0

        largo_nul = Llimpio.shape[0]

        while i < largo_nul:

            if i+9 < largo_nul:
                Lmien[k] = numpy.mean(Llimpio[i:i+5])
                Fmien[k] = numpy.median(Flimpio[i:i+5])
                i = i+5
                k = k+1
            else:
                Lmien[k] = numpy.mean(Llimpio[i:largo_nul])
                Fmien[k] = numpy.median(Flimpio[i:largo_nul])
                i = i+9
                k = k+1

        largo_fil = k
        Llimpio = Lmien[0:largo_fil]
        Flimpio = Fmien[0:largo_fil]

        Lfilter[o] = list(Llimpio)
        Ffilter[o] = list(Flimpio)

        esto[0,o] = list(Llimpio)
        esto[1,o] = list(Flimpio)

        """
        no considero strong lines:
        """
        rellenado = False

        j = 0
        while j<strli.shape[0]:

            I = numpy.where( (Llimpio > strli[j]) & (Llimpio < strlf[j]) )[0]
            Flimpio[I]=-1

            if I.shape[0] > 0:
                rellenado == True

            j=j+1

        """
        completo flujo de strong lines con ajuste lineal
        """
        if rellenado:
            Flimpio = FunNorm.Rell(Llimpio.astype("double"),Flimpio.astype("double"),Flimpio.shape[0])

        """
        divido cada orden en 3 y ajusto una recta
        """

        can = int(Llimpio.shape[0]//div)
        mm = numpy.zeros(div,float)
        nn = numpy.zeros(div,float)

        c=0
        while c < div:

            vectorL = Llimpio[c*can:(c+1)*can]
            vectorF = Flimpio[c*can:(c+1)*can]
            para = division(vectorL,vectorF,rec)

            mm[c] = para[0][0]
            nn[c] = para[0][1]
            c = c+1

        c = 0
        FF = []
        LL = []

        while c < div:

            m = mm[c]
            n = nn[c]

            if c != div-1:
                segl = Llimpio[c*can:(c+1)*can]
                segf = segl*m+n
            elif c == div-1:
                segl = Llimpio[c*can:Llimpio.shape[0]]
                segf = segl*m+n

            LL = LL+list(segl)
            FF = FF+list(segf)
            c = c+1

        LL = numpy.array(LL)
        FF = numpy.array(FF)

        ajustep = numpy.polyfit(LL,FF,orden)
        FF = numpy.polyfit(ajustep, LL)
        """
        ajustep=optimize.leastsq(res_cur,guess2,args=(FF,LL))
        a0=ajustep[0][0]
        a1=ajustep[0][1]
        a2=ajustep[0][2]
        a3=ajustep[0][3]
        a4=ajustep[0][4]

        FF=a0+a1*LL+a2*(LL**2)+a3*(LL**3)+a4*(LL**4)
        """
        """
        considero los peaks sobre 1 sigma y ajusto polinomio de orden 4
        """

        JJ = FF.copy()
        RES = Flimpio-FF
        veva = numpy.zeros(pixeles,float)
        AJ = numpy.zeros((Flimpio.shape[0]),float)

        I = numpy.where(RES > 0.0)[0]
        veva = RES[I]

        dev = math.sqrt(numpy.mean(veva*veva))

        i=0
        while i<FF.shape[0]:

            if RES[i] > dev and RES[i] <= 4*dev:
                AJ[i] = Flimpio[i]
            elif RES[i] > 4*dev:
                AJ[i] = -1
                Flimpio[i] = FF[i]
            else:
                AJ[i] = -1

            i=i+1

        JJ = FF+dev

        if AJ[0]== -1:
            AJ[0] = numpy.median(Flimpio[0:20])
        if AJ[-1] == -1:
            AJ[-1] = numpy.median(Flimpio[-21:-1])

        FFF = FunNorm.Rell(Llimpio,AJ,Flimpio.shape[0])
        LLL = Llimpio.copy()

        ajustep = numpy.polyfit(LLL,FFF,orden)
        CON = numpy.polyfit(ajustep, LLL)
        """
        ajustep = optimize.leastsq(res_cur,guess2,args=(FFF,LLL))
        a0 = ajustep[0][0]
        a1 = ajustep[0][1]
        a2 = ajustep[0][2]
        a3 = ajustep[0][3]
        a4 = ajustep[0][4]

        CON = a0+a1*LLL+a2*(LLL**2)+a3*(LLL**3)+a4*(LLL**4)
        """
        continuo[0,o] = list(LLL)
        continuo[1,o] = list(CON)

        laro[o] = LLL.shape[0]

        esto[0,o] = list(Llimpio)
        esto[1,o] = list(Flimpio)

        o=o+1

    """
    itero para subir el continuo
    """

    #print("iterando para subir el continuo")

    j = 0
    while j < 20:

        o = 0
        while o < ordenes:

            ya = numpy.array(continuo[1,o])
            RESI = numpy.array(esto[1,o])-ya
            larg = laro[o]

            loco=numpy.array(esto[0,o])
            foco=numpy.array(esto[1,o])

            I = numpy.where(RESI > 0)[0]

            ya[I] = foco[I]


            ajustec  = numpy.polyfit(loco,ya)
            aji = numpy.polyval(ajustec,loco[0:larg])
            invor = numpy.arange(orden+1)[::-1]
            for orde in invor:
                coefs[o,orde] = invor[orde]
            """
            ajustec = optimize.leastsq(res_cur,guess2,args=(ya,loco))
            A0 = ajustec[0][0]
            A1 = ajustec[0][1]
            A2 = ajustec[0][2]
            A3 = ajustec[0][3]
            A4 = ajustec[0][4]
            coefs[o,0]=A0
            coefs[o,1]=A1
            coefs[o,2]=A2
            coefs[o,3]=A3
            coefs[o,4]=A4
            aji = A0+A1*(loco[0:larg]**1)+A2*(loco[0:larg]**2)+A3*(loco[0:larg]**3)+A4*(loco[0:larg]**4)
            """
            continuo[1,o] = list(aji)

            o=o+1
        j=j+1



    """
    perfecciono bordes del ajuste con pendiente de orden anterior y posterior
    """

    #print("arreglando bordes de cada orden")

    o = 0
    while o < ordenes-1:

        larg = laro[o]
        longitud = numpy.array(continuo[0,o])
        upL = numpy.array(continuo[0,o+1])
        upF = numpy.array(continuo[1,o+1])

        lami = longitud[0]
        lamf = lami+5

        I = numpy.where( (upL >= lami) & (upL <= lamf))[0]

        if len(I)>4:

            veL = upL[I]
            veF = upF[I]

            ajuster = optimize.leastsq(res_lin,guess,args=(veF,veL))
            pen = ajuster[0][0]

            j = 0

            while j < 5:

                flujo = numpy.array(continuo[1,o])
                I = numpy.where( longitud >= lamf)[0][0]
                x = longitud[I]
                y = flujo[I]

                coe = y-pen*x

                flujo[0:I] = longitud[0:I]*pen+coe

                ajustec = numpy.polyfit(longitud,flujo,orden)
                tri = numpy.polyval(ajustec, longitud)

                """
                ajustec = optimize.leastsq(res_cur,guess2,args=(flujo,longitud))
                A0 = ajustec[0][0]
                A1 = ajustec[0][1]
                A2 = ajustec[0][2]
                A3 = ajustec[0][3]
                A4 = ajustec[0][4]

                tri=A0+A1*longitud+A2*(longitud**2)+A3*(longitud**3)+A4*(longitud**4)
                """
                continuo[1,o] = list(tri)

                j = j+1


            if o == 0:

                AJUSTE_FINAL = A0+A1*(L[o]**1)+A2*(L[o]**2)+A3*(L[o]**3)+A4*(L[o]**4)
                NORMALIZADO[o] = list(F[o]/AJUSTE_FINAL)

        else:
            AJUSTE_FINAL = numpy.polyval(coefs[o],L[o])
            #AJUSTE_FINAL = coefs[o,0]+coefs[o,1]*(L[o]**1)+coefs[o,2]*(L[o]**2)+coefs[o,3]*(L[o]**3)+coefs[o,4]*(L[o]**4)
            NORMALIZADO[o] = list(F[o]/AJUSTE_FINAL)

        o=o+1

    o = ordenes-1
    while o > 0:

        longitud=numpy.array(continuo[0,o])
        downL=numpy.array(continuo[0,o-1])
        downF=numpy.array(continuo[1,o-1])
        larg=laro[o]
        lamf=longitud[larg-1]
        lami=lamf-5

        I = numpy.where((downL>=lami) & (downL<=lamf))[0]
        if len(I)>4:
            veL = downL[I]
            veF = downF[I]

            ajuster = optimize.leastsq(res_lin,guess,args=(veF,veL))
            pen = ajuster[0][0]

            j = 0
            while j < 3:

                flujo = numpy.array(continuo[1,o])
                I = numpy.where(longitud<=lami)[0][-1]
                x = longitud[I]
                y = flujo[I]

                coe = y-pen*x

                flujo[I:] = longitud[I:]*pen+coe
                ajustec = numpy.polyfit(longitud,flujo,orden)
                tri = numpy.polyval(ajustec, longitud)
                """
                ajustec = optimize.leastsq(res_cur,guess2,args=(flujo,longitud))
                A0 = ajustec[0][0]
                A1 = ajustec[0][1]
                A2 = ajustec[0][2]
                A3 = ajustec[0][3]
                A4 = ajustec[0][4]

                tri=A0+A1*(longitud[:]**1)+A2*(longitud[:]**2)+A3*(longitud[:]**3)+A4*(longitud[:]**4)
                """
                continuo[1,o]=list(tri)

                j=j+1
            AJUSTE_FINAL = numpy.polyval( ajustec,L[o] )
            #AJUSTE_FINAL = A0+A1*(L[o]**1)+A2*(L[o]**2)+A3*(L[o]**3)+A4*(L[o]**4)

            NORMALIZADO[o] = list(F[o]/AJUSTE_FINAL)

        else:
            AJUSTE_FINAL = numpy.polyval(coefs[o],L[o])
            #AJUSTE_FINAL = coefs[o,0]+coefs[o,1]*(L[o]**1)+coefs[o,2]*(L[o]**2)+coefs[o,3]*(L[o]**3)+coefs[o,4]*(L[o]**4)
            NORMALIZADO[o] = list(F[o]/AJUSTE_FINAL)
        o=o-1

    AAAA = numpy.zeros(ordenes,list)
    BBBB = numpy.zeros(ordenes,list)

    o = 0
    while o < ordenes:

        AAAA[o] = numpy.array(NORMALIZADO[o])
        BBBB[o]=L[o]

        o = o+1;

    return BBBB,AAAA;

def NORM2(L,F):

    ordenes = L.shape[0]
    pixeles = L.shape[1]

    guess = [-0.1,1.0]
    guess2 = [1,0.1,0.1,0.1,0.1]

    div = 4
    rec = 20

    para = numpy.zeros( (2,div),float )
    continuo = numpy.zeros( (2,ordenes),list )
    strl = numpy.zeros( (6,2),float )
    laro = numpy.zeros( ordenes,int )

    #strli,strlf = numpy.loadtxt('/home/rabrahm/Desktop/corr2/strong_lines.dat',dtype=None,usecols=(0,1),unpack=True)
    strli,strlf = numpy.loadtxt('../utils/Continuum/strong_lines.dat',dtype=None,usecols=(0,1),unpack=True)

    esto = numpy.zeros( (2,ordenes),list )
    ajent = numpy.zeros( (ordenes,pixeles),float )
    unos = numpy.zeros(pixeles,float)
    NORMALIZADO = numpy.zeros(ordenes,list)
    unos = unos+1.0

    Lfilter = numpy.zeros(ordenes,list)
    Ffilter = numpy.zeros(ordenes,list)
    coefs = numpy.zeros([ordenes,5])

    o=0
    while o<ordenes:

        Lmien = numpy.zeros(pixeles,float)
        Fmien = numpy.zeros(pixeles,float)

        """
        elemino flujos nulos
        """

        I = numpy.where( F[o] != 0.0 )[0]
        Llimpio = L[o][I]
        Flimpio = F[o][I]

        """
        hago filtro de 5 pixeles
        """

        i = 0
        k = 0

        largo_nul = Llimpio.shape[0]

        while i < largo_nul:

            if i+9 < largo_nul:
                Lmien[k] = numpy.mean(Llimpio[i:i+5])
                Fmien[k] = numpy.median(Flimpio[i:i+5])
                i = i+5
                k = k+1
            else:
                Lmien[k] = numpy.mean(Llimpio[i:largo_nul])
                Fmien[k] = numpy.median(Flimpio[i:largo_nul])
                i = i+9
                k = k+1

        largo_fil = k
        Llimpio = Lmien[0:largo_fil]
        Flimpio = Fmien[0:largo_fil]

        Lfilter[o] = list(Llimpio)
        Ffilter[o] = list(Flimpio)

        esto[0,o] = list(Llimpio)
        esto[1,o] = list(Flimpio)

        """
        no considero strong lines:
        """
        rellenado = False

        j = 0
        while j<strli.shape[0]:

            I = numpy.where( (Llimpio > strli[j]) & (Llimpio < strlf[j]) )[0]
            Flimpio[I]=-1

            if I.shape[0] > 0:
                rellenado == True

            j=j+1

        """
        completo flujo de strong lines con ajuste lineal
        """
        if rellenado:
            Flimpio = FunNorm.Rell(Llimpio.astype("double"),Flimpio.astype("double"),Flimpio.shape[0])

        """
        divido cada orden en 3 y ajusto una recta
        """

        can = int(Llimpio.shape[0]//div)
        mm = numpy.zeros(div,float)
        nn = numpy.zeros(div,float)

        c=0
        while c < div:

            vectorL = Llimpio[c*can:(c+1)*can]
            vectorF = Flimpio[c*can:(c+1)*can]
            para = division(vectorL,vectorF,rec)

            mm[c] = para[0][0]
            nn[c] = para[0][1]
            c = c+1

        c = 0
        FF = []
        LL = []

        while c < div:

            m = mm[c]
            n = nn[c]

            if c != div-1:
                segl = Llimpio[c*can:(c+1)*can]
                segf = segl*m+n
            elif c == div-1:
                segl = Llimpio[c*can:Llimpio.shape[0]]
                segf = segl*m+n

            LL = LL+list(segl)
            FF = FF+list(segf)
            c = c+1

        LL = numpy.array(LL)
        FF = numpy.array(FF)

        ajustep=optimize.leastsq(res_cur,guess2,args=(FF,LL))
        a0=ajustep[0][0]
        a1=ajustep[0][1]
        a2=ajustep[0][2]
        a3=ajustep[0][3]
        a4=ajustep[0][4]

        FF=a0+a1*LL+a2*(LL**2)+a3*(LL**3)+a4*(LL**4)

        """
        considero los peaks sobre 1 sigma y ajusto polinomio de orden 4
        """

        JJ = FF.copy()
        RES = Flimpio-FF
        veva = numpy.zeros(pixeles,float)
        AJ = numpy.zeros((Flimpio.shape[0]),float)

        I = numpy.where(RES > 0.0)[0]
        veva = RES[I]

        dev = math.sqrt(numpy.mean(veva*veva))

        i=0
        while i<FF.shape[0]:

            if RES[i] > dev and RES[i] <= 4*dev:
                AJ[i] = Flimpio[i]
            elif RES[i] > 4*dev:
                AJ[i] = -1
                Flimpio[i] = FF[i]
            else:
                AJ[i] = -1

            i=i+1

        JJ = FF+dev

        if AJ[0]== -1:
            AJ[0] = numpy.median(Flimpio[0:20])
        if AJ[-1] == -1:
            AJ[-1] = numpy.median(Flimpio[-21:-1])

        FFF = FunNorm.Rell(Llimpio,AJ,Flimpio.shape[0])
        LLL = Llimpio.copy()

        ajustep = optimize.leastsq(res_cur,guess2,args=(FFF,LLL))
        a0 = ajustep[0][0]
        a1 = ajustep[0][1]
        a2 = ajustep[0][2]
        a3 = ajustep[0][3]
        a4 = ajustep[0][4]

        CON = a0+a1*LLL+a2*(LLL**2)+a3*(LLL**3)+a4*(LLL**4)

        continuo[0,o] = list(LLL)
        continuo[1,o] = list(CON)

        laro[o] = LLL.shape[0]

        esto[0,o] = list(Llimpio)
        esto[1,o] = list(Flimpio)

        o=o+1

    """
    itero para subir el continuo
    """

    #print("iterando para subir el continuo")

    j = 0
    while j < 20:

        o = 0
        while o < ordenes:

            ya = numpy.array(continuo[1,o])
            RESI = numpy.array(esto[1,o])-ya
            larg = laro[o]

            loco=numpy.array(esto[0,o])
            foco=numpy.array(esto[1,o])

            I = numpy.where(RESI > 0)[0]

            ya[I] = foco[I]

            ajustec = optimize.leastsq(res_cur,guess2,args=(ya,loco))
            A0 = ajustec[0][0]
            A1 = ajustec[0][1]
            A2 = ajustec[0][2]
            A3 = ajustec[0][3]
            A4 = ajustec[0][4]
            coefs[o,0]=A0
            coefs[o,1]=A1
            coefs[o,2]=A2
            coefs[o,3]=A3
            coefs[o,4]=A4
            aji = A0+A1*(loco[0:larg]**1)+A2*(loco[0:larg]**2)+A3*(loco[0:larg]**3)+A4*(loco[0:larg]**4)

            continuo[1,o] = list(aji)

            o=o+1
        j=j+1



    """
    perfecciono bordes del ajuste con pendiente de orden anterior y posterior
    """

    #print("arreglando bordes de cada orden")

    o = 0
    while o < ordenes-1:

        larg = laro[o]
        longitud = numpy.array(continuo[0,o])
        upL = numpy.array(continuo[0,o+1])
        upF = numpy.array(continuo[1,o+1])

        lami = longitud[0]
        lamf = lami+5

        I = numpy.where( (upL >= lami) & (upL <= lamf))[0]

        if len(I)>4:

            veL = upL[I]
            veF = upF[I]

            ajuster = optimize.leastsq(res_lin,guess,args=(veF,veL))
            pen = ajuster[0][0]

            j = 0

            while j < 5:

                flujo = numpy.array(continuo[1,o])
                I = numpy.where( longitud >= lamf)[0][0]
                x = longitud[I]
                y = flujo[I]

                coe = y-pen*x

                flujo[0:I] = longitud[0:I]*pen+coe

                ajustec = optimize.leastsq(res_cur,guess2,args=(flujo,longitud))
                A0 = ajustec[0][0]
                A1 = ajustec[0][1]
                A2 = ajustec[0][2]
                A3 = ajustec[0][3]
                A4 = ajustec[0][4]

                tri=A0+A1*longitud+A2*(longitud**2)+A3*(longitud**3)+A4*(longitud**4)

                continuo[1,o] = list(tri)

                j = j+1


            if o == 0:

                AJUSTE_FINAL = A0+A1*(L[o]**1)+A2*(L[o]**2)+A3*(L[o]**3)+A4*(L[o]**4)
                NORMALIZADO[o] = list(F[o]/AJUSTE_FINAL)

        else:
            AJUSTE_FINAL = coefs[o,0]+coefs[o,1]*(L[o]**1)+coefs[o,2]*(L[o]**2)+coefs[o,3]*(L[o]**3)+coefs[o,4]*(L[o]**4)
            NORMALIZADO[o] = list(F[o]/AJUSTE_FINAL)

        o=o+1

    o = ordenes-1
    while o > 0:

        longitud=numpy.array(continuo[0,o])
        downL=numpy.array(continuo[0,o-1])
        downF=numpy.array(continuo[1,o-1])
        larg=laro[o]
        lamf=longitud[larg-1]
        lami=lamf-5

        I = numpy.where((downL>=lami) & (downL<=lamf))[0]
        if len(I)>4:
            veL = downL[I]
            veF = downF[I]

            ajuster = optimize.leastsq(res_lin,guess,args=(veF,veL))
            pen = ajuster[0][0]

            j = 0
            while j < 3:

                flujo = numpy.array(continuo[1,o])
                I = numpy.where(longitud<=lami)[0][-1]
                x = longitud[I]
                y = flujo[I]

                coe = y-pen*x

                flujo[I:] = longitud[I:]*pen+coe

                ajustec = optimize.leastsq(res_cur,guess2,args=(flujo,longitud))
                A0 = ajustec[0][0]
                A1 = ajustec[0][1]
                A2 = ajustec[0][2]
                A3 = ajustec[0][3]
                A4 = ajustec[0][4]

                tri=A0+A1*(longitud[:]**1)+A2*(longitud[:]**2)+A3*(longitud[:]**3)+A4*(longitud[:]**4)
                continuo[1,o]=list(tri)

                j=j+1

            AJUSTE_FINAL = A0+A1*(L[o]**1)+A2*(L[o]**2)+A3*(L[o]**3)+A4*(L[o]**4)

            NORMALIZADO[o] = list(F[o]/AJUSTE_FINAL)

        else:
            AJUSTE_FINAL = coefs[o,0]+coefs[o,1]*(L[o]**1)+coefs[o,2]*(L[o]**2)+coefs[o,3]*(L[o]**3)+coefs[o,4]*(L[o]**4)
            NORMALIZADO[o] = list(F[o]/AJUSTE_FINAL)
        o=o-1

    AAAA = numpy.zeros(ordenes,list)
    BBBB = numpy.zeros(ordenes,list)

    o = 0
    while o < ordenes:

        AAAA[o] = numpy.array(NORMALIZADO[o])
        BBBB[o]=L[o]

        o = o+1;

    return BBBB,AAAA;

def NORM_single(L, F, orden = 5):

    warnings.simplefilter('ignore', numpy.RankWarning)

    pixeles = L.shape[0]
    guess = [-0.1,1.0]
    guess2 = [1,0.1,0.1,0.1,0.1]

    div = 4
    rec = 20
    nnn = orden
    para = numpy.zeros( (2,div),float )
    #strli,strlf = numpy.loadtxt('/data/echelle/ecpipe/Continuum/strong_lines.dat',dtype=None,usecols=(0,1),unpack=True)
    strli = numpy.array([3965.0,3930.0])
    strlf = numpy.array([3975.0,3940.0])
    unos = numpy.zeros(pixeles,float) + 1.0

    Lmien = numpy.zeros(pixeles,float)
    Fmien = numpy.zeros(pixeles,float)

    """
    elemino flujos nulos
    """

    I = numpy.where( F != 0.0)[0]
    Llimpio = L[I]
    Flimpio = F[I]

    """
    calculo mediana con 5 pixeles
    """

    i = 0
    k = 0

    largo_nul = Llimpio.shape[0]

    while i < largo_nul:
        if i+9 < largo_nul:
            Lmien[k] = numpy.mean(Llimpio[i:i+5])
            Fmien[k] = numpy.median(Flimpio[i:i+5])
            i = i+5
            k = k+1
        else:
            Lmien[k] = numpy.mean(Llimpio[i:largo_nul])
            Fmien[k] = numpy.median(Flimpio[i:largo_nul])
            i = i+9
            k = k+1

    largo_fil = k
    Llimpio = Lmien[0:largo_fil]
    Flimpio = Fmien[0:largo_fil]

    """
    no considero strong lines:
    """

    rellenado = False

    j = 0
    while j<strli.shape[0]:

        I = numpy.where( (Llimpio > strli[j]) & (Llimpio < strlf[j]) )[0]
        Flimpio[I]=-1

        if I.shape[0]>0:
            rellenado == True

        j=j+1

    """
    completo flujo de strong lines con ajuste lineal
    """
    if rellenado:
        #print('!!!!!!!!!!!!!!!!!!!')
        Flimpio = FunNorm.Rell(Llimpio.astype("double"),Flimpio.astype("double"),Flimpio.shape[0])
        #print(sys.getrefcount(Flimpio))

    """
    divido cada orden en 3 y ajusto una recta
    """

    can = int(Llimpio.shape[0]//div)
    mm = numpy.zeros(div,float)
    nn = numpy.zeros(div,float)

    c=0
    while c < div:

        vectorL = Llimpio[c*can:(c+1)*can+1]
        vectorF = Flimpio[c*can:(c+1)*can+1]
        para = division(vectorL,vectorF,rec)

        mm[c] = para[0][0]
        nn[c] = para[0][1]
        c = c+1

    c = 0
    FF = []
    LL = []

    while c < div:

        m = mm[c]
        n = nn[c]

        if c != div-1:
            segl = Llimpio[c*can:(c+1)*can]
            segf = segl*m+n
        elif c == div-1:
            segl = Llimpio[c*can:Llimpio.shape[0]]
            segf = segl*m+n

        LL = LL+list(segl)
        FF = FF+list(segf)
        c = c+1

    LL = numpy.array(LL)
    FF = numpy.array(FF)
    ajustep = numpy.polyfit(LL,FF,nnn)
    FF = numpy.polyval(ajustep,LL)
    """
    ajustep=optimize.leastsq(res_cur,guess2,args=(FF,LL))
    a0=ajustep[0][0]
    a1=ajustep[0][1]
    a2=ajustep[0][2]
    a3=ajustep[0][3]
    a4=ajustep[0][4]

    FF=a0+a1*LL+a2*(LL**2)+a3*(LL**3)+a4*(LL**4)
    """
    """
    considero los peaks sobre 1 sigma y ajusto polinomio de orden 4
    """

    JJ = FF.copy()
    RES = Flimpio-FF
    veva = numpy.zeros(pixeles,float)
    AJ = numpy.zeros((Flimpio.shape[0]),float)

    WI = numpy.where(RES < 0.0)[0]
    devm =  math.sqrt(numpy.mean(RES[WI]*RES[WI]))

    I = numpy.where(RES > 0.0)[0]
    veva = RES[I]

    dev = math.sqrt(numpy.mean(veva*veva))

    i=0
    while i<FF.shape[0]:

        if RES[i] > -devm and RES[i] <= 4*dev:
            AJ[i] = Flimpio[i]
        elif RES[i] > 4*dev:
            AJ[i] = -1
            Flimpio[i] = FF[i]
        else:
            AJ[i] = -1

        i=i+1

    JJ = FF+dev

    if AJ[0 ]== -1:
        AJ[0] = numpy.median(Flimpio[0:20])
    if AJ[-1] == -1:
        AJ[-1] = numpy.median(Flimpio[-21:-1])

    FFF = FunNorm.Rell(Llimpio,AJ,Flimpio.shape[0])

    LLL = Llimpio.copy()
    ajustep = numpy.polyfit(LLL,FFF,nnn)
    CON = numpy.polyval(ajustep,LLL)
    """
    ajustep = optimize.leastsq(res_cur,guess2,args=(FFF,LLL))
    a0 = ajustep[0][0]
    a1 = ajustep[0][1]
    a2 = ajustep[0][2]
    a3 = ajustep[0][3]
    a4 = ajustep[0][4]

    CON = a0+a1*LLL+a2*(LLL**2)+a3*(LLL**3)+a4*(LLL**4)
    """
    """
    itero para subir el continuo
    """

    #print("iterando para subir el continuo")

    j = 0
    while j < 10:


        RESI = Flimpio-CON
        larg = Flimpio.shape[0]

        loco = Llimpio.copy()
        foco = Flimpio.copy()

        I = numpy.where(RESI > 0.0)
        CON[I] = foco[I]

        ajustec = numpy.polyfit(loco,CON,nnn)
        aji = numpy.polyval(ajustec,loco)

        """
        ajustec = optimize.leastsq(res_cur,guess2,args=(CON,loco))
        A0 = ajustec[0][0]
        A1 = ajustec[0][1]
        A2 = ajustec[0][2]
        A3 = ajustec[0][3]
        A4 = ajustec[0][4]
        aji = A0+A1*(loco[0:larg]**1)+A2*(loco[0:larg]**2)+A3*(loco[0:larg]**3)+A4*(loco[0:larg]**4)
        """

        CON = aji.copy()


        j=j+1

    I = numpy.where( (Llimpio > Llimpio[0]+5.0) & (Llimpio < Llimpio[0]+10.0) )[0]
    if len(I) > 2:
        ajl = numpy.polyfit(Llimpio[I],aji[I],1)
        pendi = ajl[0]
        """
        ajl = optimize.leastsq(res_lin,guess,args=(aji[I],Llimpio[I]))
        pendi = ajl[0][0]
        """
        coefi = aji[I[0]]-pendi*Llimpio[I[0]]
        aj2 = aji.copy()
        aj2[0:I[0]] =  Llimpio[0:I[0]]*pendi + coefi
    else:
        aj2 = aji.copy()

    I = numpy.where( (Llimpio > Llimpio[-1]-10.0) & (Llimpio < Llimpio[-1]-5.0) )[0]
    if len(I) > 2:
        ajl = numpy.polyfit(Llimpio[I],aji[I],1)
        pendi = ajl[0]
        """
        ajl = optimize.leastsq(res_lin,guess,args=(aji[I],Llimpio[I]))
        pendi = ajl[0][0]
        """
        coefi = aji[I[-1]]-pendi*Llimpio[I[-1]]
        aj2[I[-1]:] =  Llimpio[I[-1]:]*pendi + coefi
    else:
        aj2 = aji.copy()
    ajustec = numpy.polyfit(Llimpio,aj2,nnn)
    ajf = numpy.polyval(ajustec,L)
    """
    ajustec = optimize.leastsq(res_cur,guess2,args=(aj2,Llimpio))
    A0 = ajustec[0][0]
    A1 = ajustec[0][1]
    A2 = ajustec[0][2]
    A3 = ajustec[0][3]
    A4 = ajustec[0][4]

    ajf = A0+A1*(L**1)+A2*(L**2)+A3*(L**3)+A4*(L**4)
    """
    NF = F/ajf

    gc.collect()
    #print(sys.getrefcount(para))
    #print(sys.getrefcount(unos))
    #print(sys.getrefcount(Lmien))
    return ajustec



from astropy.io import fits as pyfits
import scipy
import glob
import os
import sys

from ceres3.utils import globalutils as GLOBALutils

import numpy as np
import matplotlib.pyplot as plt


def FileClassify(dir, log,binning):
    """
    
    Classifies all files in a directory and writes a night log of science images

    """

    # define output lists
    thars      = []
    biases     = []
    quartzB    = []
    quartzR    = []
    science    = []
    obnames    = []
    exptimes   = []
    thar_dates = []

    f = open(log,'w')

    bad_files = []
    if os.access(dir+'bad_files.txt',os.F_OK):
        bf = open(dir+'bad_files.txt')
        linesbf = bf.readlines()
        for line in linesbf:
                bad_files.append(dir+line[:-1])
        bf.close()
    
    all_files = glob.glob(dir+"/*fits")
    for archivo in all_files:
        dump = False
        for bf in bad_files:
                if archivo == bf:
                        dump = True
                        break
        h = pyfits.open(archivo)
        if dump == False and h[0].header['INSTRUME'] == 'echelle':
                if h[0].header['IMAGETYP'] == 'object':
                    science.append(archivo)
                    obname  = h[0].header['OBJNAME']
                    ra      = h[0].header['RA']
                    delta   = h[0].header['DEC']
                    airmass = h[0].header['AIRMASS']
                    texp    = h[0].header['EXPTIME']
                    date    = h[0].header['DATE-OBS'][:10]
                    hour    = h[0].header['DATE-OBS'][11:]
                    obnames.append( obname )
                    exptimes.append( texp ) 
                    line = "%-15s %10s %10s %8.2f %4.2f %8s %8s %s\n" % (obname, ra, delta, texp, airmass, date, hour, archivo)
                    f.write(line)
                elif h[0].header['IMAGETYP'] == 'zero':
                    biases.append(archivo)
                elif h[0].header['IMAGETYP'] == 'flat':
                    if h[0].header['FILTER'] == 'Blue':
                        quartzB.append(archivo)
                    elif h[0].header['FILTER'] == 'Open':
                        quartzR.append(archivo)
                elif h[0].header['IMAGETYP'] == 'comp':
                    thars.append(archivo)
                    mjd, mjd0 = mjd_fromheader(h)
                    thar_dates.append( mjd )
      
    f.close()

    return biases, quartzB, quartzR, science, thars, thar_dates, obnames, exptimes

def mjd_fromheader(h):
    """
    return modified Julian date from header
    """
    
    datetu = h[0].header['DATE-OBS'][:10]
    ut     = h[0].header['DATE-OBS'][11:]
    mjd0,mjd,i = GLOBALutils.iau_cal2jd(int(datetu[0:4]),int(datetu[5:7]),int(datetu[8:]))
    ut = float(ut[:2]) + float(ut[3:5])/60. + float(ut[6:])/3600.
    mjd_start = mjd + ut/24.0
    secinday = 24*3600.0
    fraction = .5
    texp     = h[0].header['EXPTIME'] #sec
    mjd = mjd_start + (fraction * texp) / secinday

    return mjd, mjd0

def bad_col_corr(dat):
        mat = np.array([[80  ,81  ,199 ,2048],\
                        [240 ,241 ,530 ,2048],\
                        [469 ,470 ,1555,1666],\
                        [547 ,548 ,1690,1703],\
                        [632 ,634 ,1985,1962],\
                        [766 ,767 ,782 ,1980],\
                        [832 ,833 ,1850,1906],\
                        [960 ,961, 1590,1871],\
                        [1263,1264,1772,1816],\
                        [1334,1335,1454,2048],\
                        [1361,1362,1885,1924],\
                        [1395,1396,1906,1955],\
                        [1460,1461,488 ,2048],\
                        [1477,1478,758 ,2048],\
                        [1510,1511,771 ,2048],\
                        [1581,1582,1398,1763],\
                        [1629,1630,420 ,2048],\
                        [1661,1662,643 ,1980]])
        for bd in mat:
                new = .5*( dat[bd[2]:bd[3], bd[0]-1] + dat[bd[2]:bd[3],bd[1]] )
                bci = bd[0]
                while bci < bd[1]:
                        dat[ bd[2]:bd[3], bci ] = new
                        bci += 1 
        return dat

def MedianCombine(ImgList, bias = 0.):
    """
    Median combine a list of images
    """
    n = len(ImgList)
    if n == 0:
        print("\t\tWarning: 0 biases")
        return 0, 0, 0
    h = pyfits.open(ImgList[0])[0]
    d = h.data
    d = OverscanTrim(d)
    d = bad_col_corr(d)
    d -= bias
    
    factor = 1.25
    if (n < 3):
        factor = 1

    ronoise = factor * h.header['RDNOISE'] / np.sqrt(n)
    gain    = h.header['GAIN']

    
    if (n == 1):
        return d, ronoise, gain
    else:
        for i in range(n-1):
            h = pyfits.open(ImgList[i+1])[0]
            #print ImgList[i+1], h.data.shape
            d2 = OverscanTrim(h.data)
            d2 = bad_col_corr(d2)
            d = np.dstack( (d ,d2 - bias ))
        return np.median(d,axis=2), ronoise, gain

def OverscanTrim(d):
    """
    Overscan correct and Trim a refurbished APO image
    """
    ov = d[:20,21:-59]
    ov = np.median(ov,axis = 0)
    newdata = d[20:,21:-59]
    overscan = np.zeros(newdata.shape)
    for i in range(overscan.shape[1]):
        overscan[:,i] = ov

    newdata = newdata - overscan
    return newdata




import pyfits
from pylab import *
import scipy
from scipy import interpolate
import sys
from ceres3.utils import globalutils as GLOBALutils

lux = 2.99792458E5
scnd = np.array([  2.14285714e-07,   7.00000000e-04,  -3.04357143e+00])

sc = pyfits.getdata('/Users/rabrahm/data/espadons/20150107_red/1772293c.spec.ob.fits.S')
f = open('/Users/rabrahm/codes/ceres/espadons/thar_list.txt','r')
lines = f.readlines()
tlines = []
names  = []
for line in lines:
	cos = line.split()
	tlines.append(float(cos[0]))
	try:
		names.append(cos[1])
	except:
		names.append('?')
tlines = np.array(tlines)
names = np.array(names)
coefs = np.loadtxt('/Users/rabrahm/codes/ceres/espadons/wcal_ref.dat')
vels = np.arange(-500,500,10.)

nspec = np.ones((11,sc.shape[0],sc.shape[2]))
nspec[5] = sc[:,1,:]

for i in range(nspec.shape[1]):
	npix = np.arange(nspec.shape[2])
	wav  = np.polyval(coefs[i][1:][::-1],npix)*10
	nspec[0,i,:]  = wav
	spl           = interpolate.splrep(np.arange(wav.shape[0]), wav,k=3)
	dlambda_dx    = interpolate.splev(np.arange(wav.shape[0]), spl, der=1)
	NN            = np.average(dlambda_dx)
	dlambda_dx    /= NN
	nspec[9,i,:]  = nspec[5,i,:] * (dlambda_dx ** 1)
	nspec[8,i,:]  = 100.

ml_v = tlines - 0.01
mh_v = tlines + 0.01
weight = np.ones(len(tlines))
vels, xc_full, sn, nlines_ccf, W_ccf = \
                    GLOBALutils.XCor(nspec, ml_v, mh_v, weight, 0, 1., vel_width=300,vel_step=3.,\
                                          spec_order=9,iv_order=10,sn_order=8,max_vel_rough=300)
xc_av = GLOBALutils.Average_CCF(xc_full, sn, sn_min=3., Simple=True, W=W_ccf)
#plot(vels,xc_av)
#show()
j = np.median(xc_av)
im = np.argmax(xc_av)
rv = vels[im]

vels, xc_full, sn, nlines_ccf, W_ccf = \
                    GLOBALutils.XCor(nspec, ml_v, mh_v, weight, rv, 1., vel_width=20,vel_step=0.3,\
                                          spec_order=9,iv_order=10,sn_order=8,max_vel_rough=300)
xc_av = GLOBALutils.Average_CCF(xc_full, sn, sn_min=3., Simple=True, W=W_ccf)
cent = np.sum(xc_av*vels) / np.sum(xc_av)
print(cent)
#axvline(cent)
#plot(vels,xc_av-j)
#show()
#print(gfvcdxs)
for i in range(nspec.shape[1]):
	si = str(i)
	if i<10:
		si = '0'+si
	f = open('order_'+si+'.iwdat','w')
	npix = np.arange(nspec.shape[2])
	wav  = np.polyval(coefs[i][1:][::-1],npix)*10
	wav  = wav*(1-cent/lux)
	tck = interpolate.splrep(npix,wav,k=3)
	wav = interpolate.splev(npix+np.polyval(scnd,npix),tck)

	I = np.where((tlines>wav[0]) & (tlines<wav[-1]))[0]

	tck = interpolate.splrep(wav,npix,k=3)
	lns = tlines[I]
	nms = names[I]
	plns = interpolate.splev(lns,tck)
	I = np.where((plns>50)&(plns<len(wav)-50))[0]
	lns = lns[I]
	plns = plns[I]
	nms = nms[I]
	#print(plns)
	#print(lns)
	if i ==30:
		plot(npix,sc[i,1])
		plot(plns,sc[i,1,plns.astype('int')],'ro')
		show()
	toy = True
	grp1,grp2,grp3 = [lns[0]],[plns[0]],[nms[0]]
	o=1
	while o < len(lns):
		if not toy:
			grp1,grp2,grp3 = [lns[o]],[plns[o]],[nms[o]]
			toy=True
		else:
			if plns[o]-plns[o-1]<3:
				grp1.append(lns[o])
				grp2.append(plns[o])
				grp3.append(nms[o])
			else:
				toy=False
				line = str(len(grp1))+'\t'
				for u in range(len(grp1)):
					line += str(np.around(grp2[u]))+'\t'+str(grp1[u])+'\t'
				for u in range(len(grp3)):
					line += str(grp3[u])+'\t'
				line += '\n'
				f.write(line)
				o-=1
		o+=1
	f.close()





import sys
import matplotlib
matplotlib.use("Agg") 


import matplotlib.pyplot as plt

# ecpipe modules
import os
_pkg_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
_data_dir = os.path.join(_pkg_dir, 'instruments', 'data')
_xc_masks_dir = os.path.join(_pkg_dir, 'utils', 'data', 'xc_masks') + '/'

baryc_dir = ''
ephemeris = 'DEc403'

from ceres3.utils import continuum
from ceres3.utils import correlation
from ceres3.ext import Marsh
from ceres3.instruments import cafeutils
from ceres3.utils import globalutils as GLOBALutils

# other useful modules
from astropy.io import fits as pyfits
import pickle
import numpy as np
import scipy
import scipy.interpolate
from math import radians as rad
import argparse
import warnings
warnings.filterwarnings("ignore")

import ephem
from ceres3.utils import jplephem
from matplotlib.backends.backend_pdf import PdfPages

import statsmodels.api as sm
lowess = sm.nonparametric.lowess

parser = argparse.ArgumentParser()
parser.add_argument('directorio')
parser.add_argument('-o2do',default='all')
parser.add_argument('-just_extract', action="store_true", default=False)
parser.add_argument('-do_class', action="store_true", default=False)
parser.add_argument('-avoid_plot', action="store_true", default=False)
parser.add_argument('-npools', default=1)
parser.add_argument('-reffile',default='default')
parser.add_argument('-dirout',default='default')

args = parser.parse_args()
DoClass     = args.do_class
avoid_plot  = args.avoid_plot
dirin       = args.directorio
object2do   = args.o2do
JustExtract = args.just_extract
npools      = int(args.npools)
reffile     = args.reffile
dirout      = args.dirout

if dirin[-1] != '/':
    dirin = dirin + '/'

if dirout == 'default':
    dirout = dirin[:-1]+'_red/'

if not os.access(dirout,os.F_OK):
    os.makedirs(dirout, exist_ok=True)
if os.access(dirout+'proc',os.F_OK):
    os.system('rm -r '+dirout+'proc')
os.makedirs(dirout+'proc', exist_ok=True)

f_res = open(dirout+'proc/'+'results.txt','w')

if reffile == 'default':
    reffile = dirin+'reffile.txt'

####### GLOBAL VARIABLES #####
force_pre_process  = False
force_flat_extract = False
force_sci_extract  = False
force_thar_extract = False      
force_tharxc       = False
force_thar_wavcal  = False
force_spectral_file_build = True
force_stellar_pars = False
dumpargon          = False
minlines_glob      = 700

Inverse_m          = True
use_cheby          = True
MRMS               = 100   # max rms in m/s, global wav solution

trace_degree       = 4
Marsh_alg          = 0
ext_aperture       = 5
NSigma_Marsh       = 5
NCosmic_Marsh      = 5
S_Marsh            = 0.4
N_Marsh            = 3      # grado polinomio 
min_extract_col    = 200
max_extract_col    = 1848

if '2013' in dirin:
        min_extract_col    = 100
        max_extract_col    = 1800

ncoef_x            = 3
ncoef_m            = 8
npar_wsol = (min(ncoef_x,ncoef_m) + 1) * (2*max(ncoef_x,ncoef_m) - min(ncoef_x,ncoef_m) + 2) // 2

models_path = os.path.join(_pkg_dir, 'data', 'COELHO_MODELS', 'R_40000b/')
order_dir   = os.path.join(_data_dir, 'cafe_wavcals/')

n_useful = 80
ro0      = 60
#############################

# file containing the log
log = dirout+'night.log'

print("\n\n\tCAFE CAHA2.2m  PIPELINE\n")
print("\tRAW data is in ",dirin)
print("\tProducts of reduction will be in",dirout)
print('\n')

biases, flats, ThAr_ref, sim_sci, ThAr_ref_dates, obnames, exptimes = cafeutils.FileClassify(dirin,log)
IS = np.argsort(ThAr_ref_dates)
ThAr_ref_dates = ThAr_ref_dates[IS]
ThAr_ref = ThAr_ref[IS]

print('\tThese are all the images to proccess:')
f = open(log)
flines = f.readlines()
for line in flines:
        print('\t'+line[:-1])
print('\n')

if (     (os.access(dirout+'Flat.fits',os.F_OK) == False) or \
         (os.access(dirout+'trace.pkl',os.F_OK) == False)  or \
         (os.access(dirout+'MasterBias.fits',os.F_OK) == False)  or \
         (force_pre_process) ):
    print("\tNo previous pre-processing files or found")
    pre_process = 1
else:
    print("\tPre-processing files found, going straight to extraction")
    pre_process = 0


if (pre_process == 1):
    # median combine Biases
    print("\tGenerating Master calibration frames...")
    MasterBias, RO_bias, GA_bias = cafeutils.MedianCombine(biases)
    hdu = pyfits.PrimaryHDU( MasterBias )
    if (os.access(dirout+'MasterBias.fits',os.F_OK)):
        os.remove(dirout+'MasterBias.fits')
    hdu.writeto(dirout+'MasterBias.fits')
    print("\t\t-> Masterbias: done!")

    # median combine list of flats
    Flat, RO_fl, GA_fl = cafeutils.MedianCombine(flats, zero=dirout+'MasterBias.fits')
    hdu = pyfits.PrimaryHDU(Flat)
    if (os.access(dirout+'Flat.fits',os.F_OK)):
        os.remove(dirout+'Flat.fits')
    hdu.writeto(dirout+'Flat.fits')
    print("\t\t-> Masterflats: done!")

    print("\tTracing echelle orders...")
    c_all,nord = GLOBALutils.get_them(Flat, ext_aperture, trace_degree, maxords=-1, mode=1)
    c_all, nord = GLOBALutils.good_orders(c_all,nord,Flat.shape[0],Flat.shape[1],ext_aperture)
    if nord >= 90:
        c_all = c_all[:90]
        nord=90
    print('\t\t'+ str(nord)+' orders found.')

    # pickle traces
    trace_dict = {'c_all':c_all,
                  'nord':nord,
                  'GA_bias': GA_bias, 'RO_bias' : RO_bias,
                  'GA_fl': GA_fl, 'RO_fl': RO_fl}
    pickle.dump(trace_dict, open(dirout+"trace.pkl", 'wb'))

else:
    trace_dict = pickle.load(open(dirout+"trace.pkl", 'rb'), encoding='latin1')
    c_all = trace_dict['c_all']
    nord  = trace_dict['nord']
    # recover GA*, RO*
    GA_bias = trace_dict['GA_bias']
    RO_bias = trace_dict['RO_bias']
    GA_fl = trace_dict['GA_fl']
    RO_fl = trace_dict['RO_fl']
    # recover flats & master bias
    h = pyfits.open(dirout+'Flat.fits')
    Flat = h[0].data
    h = pyfits.open(dirout+'MasterBias.fits')
    MasterBias = h[0].data

print('\n\tExtraction of Flat calibration frames:')

P_fits      = dirout + 'P.fits'
S_flat_fits = dirout +'flat.fits'
S_flat      = np.zeros((nord, 3, Flat.shape[1]) )

if ( os.access(P_fits,os.F_OK) == False ) or \
   ( os.access(S_flat_fits,os.F_OK) == False ) or \
   (force_flat_extract):
    print("\t\tNo extracted flat object spectra found or extraction forced, extracting and saving...")

    Centers = np.zeros((len(c_all),Flat.shape[1]))
    for i in range(nord):
        Centers[i,:]=np.polyval(c_all[i,:],np.arange(len(Centers[i,:])))

    bac = GLOBALutils.get_scat(Flat,Centers,span=7)
    fl = Flat - bac
    #plot(fl[:,1000])
    #plot(np.around(Centers[:,1000]).astype('int'),fl[np.around(Centers[:,1000].astype('int')),1000],'ro')
    #show()
    #print gfd
    bacfile = dirout + 'BAC_FLAT.fits'
    if (os.access(bacfile,os.F_OK)):
            os.remove( bacfile )
    hdbac = pyfits.PrimaryHDU( bac )
    hdbac.writeto(bacfile)

    print("\t\tWill extract",nord,"orders for object fibre...")
    P = GLOBALutils.obtain_P(fl,c_all,ext_aperture,RO_fl,\
                                    GA_fl,NSigma_Marsh, S_Marsh, \
                    N_Marsh, Marsh_alg, min_extract_col,\
                    max_extract_col, npools)

    if (os.access(P_fits,os.F_OK)):
        os.remove( P_fits )
    hdu = pyfits.PrimaryHDU( P )
    hdu.writeto( P_fits )

    S_flat  = GLOBALutils.optimal_extraction(fl,P,c_all,ext_aperture,RO_fl,GA_fl,\
              S_Marsh,10*NCosmic_Marsh,min_extract_col,max_extract_col,npools)
    for i in range(nord):
        S_flat[i,1] = S_flat[i,1][::-1]
        S_flat[i,2] = S_flat[i,2][::-1]

    if (os.access(S_flat_fits,os.F_OK)):
        os.remove( S_flat_fits )
    hdu = pyfits.PrimaryHDU( S_flat )
    hdu.writeto( S_flat_fits )

else:
    print("\t\tExtracted flat object spectra found, loading...")
    P    = pyfits.getdata( P_fits )
    S_flat = pyfits.getdata( S_flat_fits )

# Normalize flat field spectra.
S_flat_n, Snorms = GLOBALutils.FlatNormalize_single( S_flat, mid = int(.5*S_flat.shape[2]) )

new_list = []
new_times = []
for fsim in sim_sci:
        h = pyfits.open(fsim)
        if object2do in h[0].header['OBJECT'] or object2do == 'all':
                new_list.append(fsim)
                mjd, mjd0 = cafeutils.mjd_fromheader2( h )
                new_times.append(mjd)

new_list  = np.array(new_list)
new_times = np.array(new_times)
IS        = np.argsort(new_times)
new_list  = new_list[IS]
new_times = new_times[IS]

I1 = np.where(ThAr_ref_dates < new_times.min())[0]
if len(I1)>0:
    I1 = I1[-1]
else:
    I1 = 0
I2 = np.where(ThAr_ref_dates > new_times.max())[0]
if len(I2)>0:
    I2 = I2[0]
else:
    I2 = len(ThAr_ref_dates)

ThAr_ref_dates = ThAr_ref_dates[I1:I2+1]
ThAr_ref = ThAr_ref[I1:I2+1]

print('\n\tExtraction of ThAr calibration frames:')

# Extract all ThAr files
for fsim in ThAr_ref:
    thar_fits        = dirout + fsim.split('/')[-1][:-4]+'spec.fits.S'
    thar_simple_fits = dirout + fsim.split('/')[-1][:-4]+'spec.simple.fits.S'

    if ( os.access(thar_simple_fits,os.F_OK) == False ) or ( force_thar_extract ):
        hthar = pyfits.open( fsim )
        dthar = cafeutils.OverscanTrim( hthar[0].data ) - MasterBias
        Centers = np.zeros((len(c_all),dthar.shape[1]))
        for i in range(nord):
            Centers[i,:]=np.polyval(c_all[i,:],np.arange(len(Centers[i,:])))
        bac = GLOBALutils.get_scat(dthar,Centers,span=7)
        sdthar = dthar - bac
        print("\t\tNo previous extraction or extraction forced for ThAr file", fsim, "extracting...")
        thar_S  = np.zeros( (nord,3,dthar.shape[1]) )
        thar_Ss = np.zeros( (nord,dthar.shape[1]) )

        tR,tG = hthar[0].header['CCDRON'],hthar[0].header['CCDGAIN']
        #print tR,tG
        thar_Ss = GLOBALutils.simple_extraction(dthar,c_all,ext_aperture,\
                                                  min_extract_col,max_extract_col,npools)
        for i in range(nord):
            thar_Ss[i] = thar_Ss[i][::-1]
        # save as fits file
        if (os.access(thar_fits,os.F_OK)):
            os.remove( thar_fits )
        if (os.access(thar_simple_fits,os.F_OK)):
            os.remove( thar_simple_fits )
            
        hdu = pyfits.PrimaryHDU( thar_S )
        hdu.writeto( thar_fits )
        hdu = pyfits.PrimaryHDU( thar_Ss )
        hdu.writeto( thar_simple_fits )
    else:
        print("\t\tThAr file", fsim, "all ready extracted, loading...")

or50 = 50
or502 = 50
maxxc = 0.
maxxc2 = 0.
thar_fits = dirout + ThAr_ref[0].split('/')[-1][:-4]+'spec.simple.fits.S'
thar_Ss   = pyfits.getdata(thar_fits)

for i in range(40,60,1):
        ccf_max, shift = GLOBALutils.cor_thar(thar_Ss[i],span=40,filename=order_dir+'order_103.iwdat')
        if ccf_max > maxxc:
                maxxc      = ccf_max
                rough_shift = shift
                or50       =  i
difo = 103 - or50
#show()
#print gfds
print("\n\tWavelength solution of ThAr calibration spectra:")

for i in range(len(ThAr_ref_dates)):
    wavsol_pkl = dirout + ThAr_ref[i].split('/')[-1][:-4]+'wavsolpars.pkl'
    hd = pyfits.getheader(ThAr_ref[i])
    gain,ron = float(hd['CCDGAIN']),float(hd['CCDRON'])
    #force_thar_wavcal = True
    if ( os.access(wavsol_pkl,os.F_OK) == False ) or (force_thar_wavcal):
        print("\t\tWorking on initial ThAr file", ThAr_ref[i])
        hthar = pyfits.open( ThAr_ref[i] )
        mjd, mjd0 = cafeutils.mjd_fromheader2( hthar )
        thar_fits = dirout + ThAr_ref[i].split('/')[-1][:-4]+'spec.simple.fits.S'
        thar_S = pyfits.getdata( thar_fits )
        lines_thar  = thar_S.copy()
        #iv_thar_ob_B     = thar_S_ob_B[:,2,:]
        c_p2w = np.zeros((n_useful,4))
        wavs,ords = [],[]
        All_Pixel_Centers = np.array([])
        All_Wavelengths   = np.array([])
        All_Orders        = np.array([])
        All_Centroids     = np.array([])
        All_Sigmas        = np.array([])
        All_Intensities   = np.array([])
        All_residuals     = np.array([])
        torder = 0
        for order in range(ro0,ro0+n_useful+1,1):
            #print order
            order_s = str(order)
            if (order < 100):
                order_s = '0' + str(order)
            thar_order_orig = lines_thar[order - difo]
            #IV              = iv_thar_ob_R[order,:]
            L = np.where(thar_order_orig != 0)[0]
            IV = 1. / (thar_order_orig / gain + (ron/gain)**2 )
            IV[L] = 0.
            wei             = np.sqrt( IV )
            thar_order      = thar_order_orig #- bkg

            coeffs_pix2wav, coeffs_pix2sigma, pixel_centers, wavelengths, rms_ms, residuals, centroids,sigmas, intensities \
                = GLOBALutils.Initial_Wav_Calibration( order_dir+'order_'+order_s+'.iwdat', thar_order, order, np.ones(len(thar_order)), rmsmax=300, minlines=10, FixEnds=False,Dump_Argon=dumpargon, Dump_AllLines=True, Cheby=use_cheby,porder=3,rough_shift=rough_shift)

            if (order == int(.5*n_useful)+ro0): 
                
                if (use_cheby):
                    Global_ZP = GLOBALutils.Cheby_eval( coeffs_pix2wav,  0.5*len(thar_order), len(thar_order) )
                else:
                    Global_ZP = np.polyval( coeffs_pix2wav, 0.0 )

            All_Pixel_Centers = np.append( All_Pixel_Centers, pixel_centers )
            All_Wavelengths   = np.append( All_Wavelengths, wavelengths )
            All_Orders        = np.append( All_Orders, np.zeros( len(pixel_centers) ) + order -ro0)
            All_Centroids     = np.append( All_Centroids, centroids)
            All_Sigmas        = np.append( All_Sigmas, sigmas)
            All_Intensities   = np.append( All_Intensities, intensities )
            All_residuals     = np.append( All_residuals, residuals )
            wavs.append(GLOBALutils.Cheby_eval( coeffs_pix2wav,  0.5*len(thar_order), len(thar_order) ))
            ords.append(torder)
            torder += 1

        p0    = np.zeros( npar_wsol )
        p0[0] =  (int(.5*n_useful)+ro0) * Global_ZP 
        p1, G_pix, G_ord, G_wav, II, rms_ms, G_res = GLOBALutils.Fit_Global_Wav_Solution(All_Pixel_Centers, All_Wavelengths,\
                                                     All_Orders, np.ones(All_Intensities.shape), p0, Cheby=use_cheby,       \
                                                     order0=ro0, ntotal=n_useful, maxrms=200, Inv=Inverse_m, minlines=200,  \
                                                     npix=len(thar_order),nx=ncoef_x,nm=ncoef_m)

        pdict = {'p1':p1,'mjd':mjd, 'G_pix':G_pix, 'G_ord':G_ord, 'G_wav':G_wav, 'II':II, 'rms_ms':rms_ms,\
                     'G_res':G_res, 'All_Centroids':All_Centroids, 'All_Wavelengths':All_Wavelengths, 'All_Orders':All_Orders, 'All_Pixel_Centers':All_Pixel_Centers, 'All_Sigmas':All_Sigmas}
        pickle.dump(pdict, open(wavsol_pkl, 'wb'))

    else:
        print("\t\tUsing previously computed wavelength solution in file",wavsol_pkl)
        pdict = pickle.load(open(wavsol_pkl, 'rb'), encoding='latin1')


wavsol_pkl = dirout + ThAr_ref[0].split('/')[-1][:-4]+'wavsolpars.pkl'
dct        = pickle.load(open(wavsol_pkl, 'rb'), encoding='latin1')
p_ref = dct['p1']
mjds_thar,shifts = [],[]

print('\n\tDetermination of instrumental drift during the night...')
temps = []
for i in range(len(ThAr_ref_dates)):
    wavsol_pkl = dirout + ThAr_ref[i].split('/')[-1][:-4]+'wavsolpars.pkl'
    hthar = pyfits.open( ThAr_ref[i] )
    npix = hthar[0].data.shape[1]
    temps.append(float(hthar[0].header['HIERARCH CAHA GEN AMBI PRES']))
    mjd, mjd0 = cafeutils.mjd_fromheader2( hthar )
    mjds_thar.append(mjd)
    dct = pickle.load(open(wavsol_pkl, 'rb'), encoding='latin1')
    p_shift, pix_centers, orders, wavelengths, I, rms_ms, residuals  = \
                    GLOBALutils.Global_Wav_Solution_vel_shift(dct['All_Pixel_Centers'], dct['All_Wavelengths'], dct['All_Orders'],\
                                                        np.ones(len(dct['All_Orders'])), p_ref, order0=ro0, npix=npix,\
                                                        Cheby=use_cheby, ntotal=n_useful, maxrms=200, Inv=Inverse_m, minlines=500, nx=ncoef_x,nm=ncoef_m)
    shifts.append(p_shift[0])

mjds_thar,shifts,temps = np.array(mjds_thar),np.array(shifts),np.array(temps)-temps[0]
shv = (1e-6*shifts)*299792458.0
if len(mjds_thar) > 1:
        tck_v = scipy.interpolate.splrep(mjds_thar,shv,k=1)
        tck_shift = scipy.interpolate.splrep(mjds_thar,shifts,k=1)
        temp = np.arange(mjds_thar[0],mjds_thar[-1],0.0001)

        mjdsh_thar = (mjds_thar - int(temp.min())) * 24.
        thour = (temp - int(temp.min())) * 24.
        thours = (new_times - int(temp.min())) * 24.
        pp = PdfPages(dirout + 'proc/ThAr_shifts.pdf')
        plot(mjdsh_thar,shv,'ro')
        plot(thour,scipy.interpolate.splev(temp,tck_v),'b')
        for i in range(len(thours)):
                axvline(x=thours[i],color ='g')
        xlabel('UT [hours]')
        ylabel('ThAr vel drift [m/s]')
        savefig(dirout + 'proc/ThAr_shifts.pdf')
        pp.savefig()
        clf()
        plot(shv,temps,'ro')
        xlabel('ThAr velocity drift [m/s]')
        ylabel('TEMP')
        pp.savefig()
        pp.close()
        clf()

        thout = np.vstack((mjdsh_thar,shv)).T
        np.savetxt(dirout + 'proc/ThAr_shifts.dat',thout)

# Does any image have a special requirement for dealing with the moonlight?
if os.access(dirin + 'moon_corr.txt', os.F_OK):
    fmoon = open(dirin + 'moon_corr.txt','r')
    moon_lns = fmoon.readlines()
    spec_moon = []
    use_moon = []
    for line in moon_lns:
        spec_moon.append(line.split()[0])
        if line.split()[1] == '0':
            use_moon.append(False)
        else:
            use_moon.append(True)
else:
    spec_moon = []
    use_moon = []

spec_moon = np.array(spec_moon)
use_moon  = np.array(use_moon)

rft,rfd = [],[]
moon_alts,moon_ills = {},{}
for fsim in new_list:

    know_moon = False
    if fsim.split('/')[-1] in spec_moon:
        I = np.where(fsim.split('/')[-1] == spec_moon)[0]
        know_moon = True
        here_moon = use_moon[I]

    print('\n')
    print("\t--> Working on image: ", fsim)
    h             = pyfits.open(fsim)
    mjd,mjd0      = cafeutils.mjd_fromheader2(h)
    ronoise, gain = float(h[0].header['CCDRON']),float(h[0].header['CCDGAIN'])

    # Object name
    obname    = h[0].header['OBJECT']
    print("\t\tObject name:",obname)

    # Open file, trim, overscan subtract and MasterBias subtract
    data        = cafeutils.b_col(h[0].data)
    data        = cafeutils.OverscanTrim( data ) - MasterBias
    drift,c_new = GLOBALutils.get_drift(data,P,c_all,pii=1024,win=10)
    P_new       = GLOBALutils.shift_P(P,drift,c_new,ext_aperture)
    #print 'ydrift:',drift

    bacfile = dirout + 'BAC_' + fsim.split('/')[-1][:-4]+'fits'
    if os.access(bacfile,os.F_OK)==False:
        Centers = np.zeros((len(c_new),data.shape[1]))
        for i in range(nord):
            Centers[i,:]=np.polyval(c_new[i,:],np.arange(len(Centers[i,:])))
        bac = GLOBALutils.get_scat(data,Centers,span=7)
        if (os.access(bacfile,os.F_OK)):
                os.remove( bacfile )
        hdbac = pyfits.PrimaryHDU( bac )
        hdbac.writeto(bacfile)
    else:
        bac = pyfits.getdata(bacfile)
    data = data - bac

    ra          = 360.*float(h[0].header['RA'])/(3600.*24)
    dec         = float(h[0].header['DEC'])/3600
    altitude    = float(h[0].header['HIERARCH CAHA TEL GEOELEV'])
    latitude    = float(h[0].header['HIERARCH CAHA TEL GEOLAT'])
    longitude   = float(h[0].header['HIERARCH CAHA TEL GEOLON'])
    epoch       = float(h[0].header['EQUINOX'])

    ra2,dec2 = GLOBALutils.getcoords(obname,mjd,filen=reffile)
    if ra2 !=0 and dec2 != 0:
        ra = ra2
        dec = dec2
    else:
        print('\t\tUsing the coordinates found in the image header.')

    iers                    = GLOBALutils.JPLiers( baryc_dir, mjd-999.0, mjd+999.0 )
    obsradius, R0           = GLOBALutils.JPLR0( latitude, altitude)
    obpos                   = GLOBALutils.obspos( longitude, obsradius, R0 )
    jplephem.set_ephemeris_dir( baryc_dir , ephemeris )
    jplephem.set_observer_coordinates( float(obpos[0]), float(obpos[1]), float(obpos[2]) )
    res         = jplephem.doppler_fraction(float(ra/15.0), float(dec), int(mjd), float(mjd%1), 1, 0.0)
    lbary_ltopo = 1.0 + res['frac'][0]
    bcvel_baryc = ( lbary_ltopo - 1.0 ) * 2.99792458E5
    print("\t\tBarycentric velocity:", bcvel_baryc)
    res  = jplephem.pulse_delay(ra/15.0, dec, int(mjd), mjd%1, 1, 0.0)
    mbjd = mjd + res['delay'][0] / (3600.0 * 24.0)

    gobs      = ephem.Observer()  
    gobs.name = h[0].header['TELESCOP']
    gobs.lat  = rad(latitude)  # lat/long in decimal degrees  
    gobs.long = rad(longitude) 
    gobs.date = h[0].header['DATE'][:10] + ' ' + h[0].header['DATE'][11:]
    mephem    = ephem.Moon()
    mephem.compute(gobs)

    Mcoo = jplephem.object_track("Moon", int(mjd), float(mjd%1), 1, 0.0)
    Mp = jplephem.barycentric_object_track("Moon", int(mjd), float(mjd%1), 1, 0.0)
    Sp = jplephem.barycentric_object_track("Sun", int(mjd), float(mjd%1), 1, 0.0)
    res  = jplephem.object_doppler("Moon", int(mjd), mjd%1, 1, 0.0)
    lunation,moon_state,moonsep,moonvel = GLOBALutils.get_lunar_props(ephem,gobs,Mcoo,Mp,Sp,res,ra,dec)
    refvel = bcvel_baryc + moonvel
    print('\t\tRadial Velocity of sacttered moonlight:',refvel)
    moon_alts.update({fsim:mephem.alt})
    moon_ills.update({fsim:lunation})

    print('\t\tExtraction:')
    sci_fits        = dirout + fsim.split('/')[-1][:-4]+'spec.fits.S'
    sci_fits_simple = dirout + fsim.split('/')[-1][:-4]+'spec.simple.fits.S'
    if ( os.access(sci_fits,os.F_OK) == False ) or ( os.access(sci_fits_simple,os.F_OK) == False ) or \
       ( force_sci_extract ):

        print("\t\t\tNo previous extraction or extraction forced for science file", fsim, "extracting...")
        sci_S  = np.zeros( (nord,3,data.shape[1]) )
        sci_Ss = np.zeros( (nord,data.shape[1]) )
        
        order = ro0

        sci_Ss = GLOBALutils.simple_extraction(data,c_new,ext_aperture,\
                                                  min_extract_col,max_extract_col,npools)
        sci_S  = GLOBALutils.optimal_extraction(data,P_new,c_new,ext_aperture,\
                                                   ronoise,gain,S_Marsh,NCosmic_Marsh,\
                                                   min_extract_col,max_extract_col,npools)
        for i in range(nord):
            sci_S[i,1,:] = sci_S[i,1,:][::-1]
            sci_S[i,2,:] = sci_S[i,2,:][::-1]
            sci_Ss[i,:]  = sci_Ss[i][::-1]
            
        if (os.access(sci_fits,os.F_OK)):
            os.remove( sci_fits )
        if (os.access(sci_fits_simple,os.F_OK)):
            os.remove( sci_fits_simple )

        hdu = pyfits.PrimaryHDU( sci_S )
        hdu.writeto( sci_fits )
        hdu = pyfits.PrimaryHDU( sci_Ss )
        hdu.writeto( sci_fits_simple )
        

    else:
        print('\t\t\t '+fsim+" has already been extracted, reading in product fits files...")
        sci_S  = pyfits.getdata( sci_fits )
        sci_Ss = pyfits.getdata( sci_fits_simple )

    fout = 'proc/' + obname + '_' + h[0].header['DATE'][:4] + h[0].header['DATE'][5:7] + h[0].header['DATE'][8:10] + '_' +'UT' + h[0].header['DATE'][11:] + '_sp.fits'
    if ( os.access(dirout+fout ,os.F_OK) == False ) or (force_spectral_file_build):
        spec = np.zeros((11, n_useful, data.shape[1]))
        hdu  = pyfits.PrimaryHDU( spec )
        hdu = GLOBALutils.update_header(hdu,'HIERARCH MJD', mjd)
        hdu = GLOBALutils.update_header(hdu,'HIERARCH MBJD', mbjd)
        hdu = GLOBALutils.update_header(hdu,'HIERARCH SHUTTER START DATE', h[0].header['DATE'][:10] )
        hdu = GLOBALutils.update_header(hdu,'HIERARCH SHUTTER START UT',  h[0].header['DATE'][11:])
        hdu = GLOBALutils.update_header(hdu,'HIERARCH TEXP (S)',h[0].header['EXPTIME'])
        hdu = GLOBALutils.update_header(hdu,'HIERARCH BARYCENTRIC CORRECTION (KM/S)', bcvel_baryc)
        hdu = GLOBALutils.update_header(hdu,'HIERARCH (LAMBDA_BARY / LAMBDA_TOPO)', lbary_ltopo)    
        hdu = GLOBALutils.update_header(hdu,'HIERARCH TARGET NAME', obname)
        hdu = GLOBALutils.update_header(hdu,'HIERARCH RA',h[0].header['RA'])
        hdu = GLOBALutils.update_header(hdu,'HIERARCH DEC',h[0].header['DEC'])
        hdu = GLOBALutils.update_header(hdu,'HIERARCH RA BARY',ra)
        hdu = GLOBALutils.update_header(hdu,'HIERARCH DEC BARY',dec)
        hdu = GLOBALutils.update_header(hdu,'HIERARCH EQUINOX',h[0].header['EQUINOX'])
        hdu = GLOBALutils.update_header(hdu,'HIERARCH OBS LATITUDE',h[0].header['HIERARCH CAHA TEL GEOLAT'])
        hdu = GLOBALutils.update_header(hdu,'HIERARCH OBS LONGITUDE',h[0].header['HIERARCH CAHA TEL GEOLON'])
        hdu = GLOBALutils.update_header(hdu,'HIERARCH OBS ALTITUDE',h[0].header['HIERARCH CAHA TEL GEOELEV'])
        hdu = GLOBALutils.update_header(hdu,'HIERARCH TARG AIRMASS START',h[0].header['AIRMASS'])
        hdu = GLOBALutils.update_header(hdu,'HIERARCH MOON VEL',refvel)

        print('\t\tWavelength calibration:')
        print("\t\t\tInstrumental drift:",(1e-6*p_shift)*299792458.0)
        # Apply new wavelength solution including barycentric correction
        equis = np.arange( data.shape[1] )
        if len(mjds_thar) > 1:
                p_shift = scipy.interpolate.splev(mjd,tck_shift)
        else:
                p_shift = 0.
        order = ro0
        ind   = 0
        while order < ro0 + n_useful:
            oss = order - difo
            m   = order
            chebs  = GLOBALutils.Calculate_chebs(equis, m, npix=data.shape[1], order0=ro0, ntotal=n_useful, Inverse=Inverse_m,nx=ncoef_x,nm=ncoef_m)
            WavSol = GLOBALutils.ToVacuum( lbary_ltopo * (1.0 + 1.0e-6*p_shift) * (1.0/float(m)) * \
                     GLOBALutils.Joint_Polynomial_Cheby(p_ref,chebs,ncoef_x,ncoef_m) )

            spec[0,ind,:] = WavSol
            spec[1,ind,:] = sci_S[oss,1]
            spec[2,ind,:] = sci_S[oss,2]
            fn  = S_flat[oss,1,:]
            L  = np.where( fn == 0 )[0]
            spec[3,ind,:] = spec[1,ind,:] / S_flat[oss,1,:]
            spec[4,ind,:] = spec[2,ind] * ( S_flat_n[oss,1,:] ** 2 )
            spec[3,ind,L] = 0.
            spec[4,ind,L] = 0.
            ind+=1
            order+=1
        ccoefs = GLOBALutils.get_cont(spec[0,:,:],spec[3,:,:])
        order = ro0
        ind   = 0
        while order < ro0 + n_useful:
            oss = order - difo
            L  = np.where( spec[1,ind] != 0 )
            spec[5,ind,:][L] = spec[3,ind][L] / np.polyval(ccoefs[ind],spec[0,ind][L])    
            ratio            = np.polyval(ccoefs[ind],spec[0,ind][L]) * Snorms[oss]
            spec[6,ind,:][L] = spec[4,ind][L] * (ratio ** 2 )
            spec[7,ind,:][L] = ratio
            spec[8,ind,:][L] = ratio * S_flat_n[oss,1][L] / np.sqrt( ratio * S_flat_n[oss,1][L] / gain + (ronoise/gain)**2 )
            spl           = scipy.interpolate.splrep(np.arange(WavSol.shape[0]), WavSol,k=3)
            dlambda_dx    = scipy.interpolate.splev(np.arange(WavSol.shape[0]), spl, der=1)
            NN            = np.average(dlambda_dx)
            dlambda_dx    /= NN
            LL = np.where(spec[5,ind] > 1 + 10. / scipy.signal.medfilt(spec[8,ind],21))[0]
            spec[5,ind,LL] = 1.

            spec[9,ind][L] = spec[5,ind][L] * (dlambda_dx[L] ** 1) 
            spec[10,ind][L] = spec[6,ind][L] / (dlambda_dx[L] ** 2)
            ind +=1
            order +=1

        if os.access(dirout + fout, os.F_OK):
                os.remove(dirout + fout)
        hdu.writeto(dirout + fout)

        if (not JustExtract):

            if DoClass:
                print('\t\tSpectral Analysis:')
                # spectral analysis
                # First, query SIMBAD with the object name
                query_success = False
                query_success,sp_type_query = GLOBALutils.simbad_query_obname(obname)
                # Now, query SIMBAD by coordinates if above not successful
                if (not query_success):
                    query_success,sp_type_query = GLOBALutils.simbad_query_coords('12:00:00','00:00:00')
                print("\t\t\tSpectral type returned by SIMBAD query:",sp_type_query)

                hdu = GLOBALutils.update_header(hdu,'HIERARCH SIMBAD SPTYP', sp_type_query)

                pars_file = dirout + fsim.split('/')[-1][:-8]+'_stellar_pars.txt'

                if os.access(pars_file,os.F_OK) == False or force_stellar_pars:
                    print("\t\t\tEstimating atmospheric parameters:")
                    Rx = np.around(1./np.sqrt(1./40000.**2 - 1./60000.**2))
                    spec2 = spec.copy()
                    for i in range(spec.shape[1]):
                        IJ = np.where(spec[5,i]!=0.)[0]
                        spec2[5,i,IJ] = GLOBALutils.convolve(spec[0,i,IJ],spec[5,i,IJ],Rx)
                    T_eff, logg, Z, vsini, vel0, ccf = correlation.CCF(spec2,model_path=models_path,npools=npools)
                    line = "%6d %4.1f %4.1f %8.1f %8.1f\n" % (T_eff,logg, Z, vsini, vel0)
                    f = open(pars_file,'w')
                    f.write(line)
                    f.close()
                       
                else:
                    print("\t\t\tAtmospheric parameters loaded from file:")
                    T_eff, logg, Z, vsini, vel0 = np.loadtxt(pars_file,unpack=True)

                print("\t\t\t\tT_eff=",T_eff,"log(g)=",logg,"Z=",Z,"vsin(i)=",vsini,"vel0",vel0)

            else:
                T_eff, logg, Z, vsini, vel0 = -999,-999,-999,-999,-999

            T_eff_epoch = T_eff
            logg_epoch  = logg
            Z_epoch     = Z
            vsini_epoch = vsini
            vel0_epoch  = vel0
            hdu = GLOBALutils.update_header(hdu,'HIERARCH TEFF', float(T_eff))
            hdu = GLOBALutils.update_header(hdu,'HIERARCH LOGG', float(logg))
            hdu = GLOBALutils.update_header(hdu,'HIERARCH Z', Z)
            hdu = GLOBALutils.update_header(hdu,'HIERARCH VSINI', vsini)
            hdu = GLOBALutils.update_header(hdu,'HIERARCH VEL0', vel0)

            print("\t\tRadial Velocity analysis:")
            # assign mask
            sp_type, mask = GLOBALutils.get_mask_reffile(obname,reffile=reffile,base=_xc_masks_dir)
            print("\t\t\tWill use",sp_type,"mask for CCF.")

            # Read in mask
            ml, mh, weight = np.loadtxt(mask,unpack=True)
            ml_v = GLOBALutils.ToVacuum( ml )
            mh_v = GLOBALutils.ToVacuum( mh )
       
            # make mask larger accounting for factor ~2 lower res in CORALIE w/r to HARPS
            av_m = 0.5*( ml_v + mh_v )
            ml_v -= (av_m - ml_v)
            mh_v += (mh_v - av_m)
            mask_hw_kms = (GLOBALutils.Constants.c/1e3) * 0.5*(mh_v - ml_v) / av_m

            #sigma_fout = stellar_pars_dir + obname + '_' +'sigma.txt'

            disp = GLOBALutils.get_disp(obname, reffile=reffile)
            if disp == 0:
                known_sigma = False
                if vsini != -999 and vsini != 0.:
                    disp = vsini
                else:
                    disp = 3.
            else:
                known_sigma = True

            mask_hw_wide = av_m * disp / (GLOBALutils.Constants.c/1.0e3)
            ml_v = av_m - mask_hw_wide
            mh_v = av_m + mask_hw_wide 

            print('\t\t\tComputing the CCF...')
            cond = True

            while (cond):
                # first rough correlation to find the minimum
                vels, xc_full, sn, nlines_ccf, W_ccf = \
                    GLOBALutils.XCor(spec, ml_v, mh_v, weight, 0, lbary_ltopo, vel_width=300,vel_step=3,\
                                          spec_order=9,iv_order=10,sn_order=8,max_vel_rough=300)
                xc_av = GLOBALutils.Average_CCF(xc_full, sn, sn_min=3.0, Simple=True, W=W_ccf)

                # Normalize the continuum of the CCF robustly with R     
                yy = scipy.signal.medfilt(xc_av,11)
                pred = lowess(yy, vels,frac=0.4,it=10,return_sorted=False)
                tck1 = scipy.interpolate.splrep(vels,pred,k=1)
                xc_av_orig = xc_av.copy()
                xc_av /= pred
                vel0_xc = vels[ np.argmin( xc_av ) ]
                rvels, rxc_av, rpred, rxc_av_orig, rvel0_xc = vels.copy(), \
                        xc_av.copy(), pred.copy(), xc_av_orig.copy(), vel0_xc

                xc_av_rough = xc_av
                vels_rough  = vels
                if disp > 30:
                        disp = 30.
                vel_width = np.maximum( 20.0, 6*disp )

                vels, xc_full, sn, nlines_ccf, W_ccf =\
                    GLOBALutils.XCor(spec, ml_v, mh_v, weight, vel0_xc, lbary_ltopo, vel_width=vel_width,vel_step=0.1,\
                                          spec_order=9,iv_order=10,sn_order=8,max_vel_rough=300)

                xc_av = GLOBALutils.Average_CCF(xc_full, sn, sn_min=3.0, Simple=True, W=W_ccf)
                pred = scipy.interpolate.splev(vels,tck1)
                xc_av /= pred
                
                if sp_type == 'M5':
                        moon_sig = 2.5
                elif sp_type == 'K5':
                        moon_sig = 3.3
                else:
                        moon_sig = 4.5

                p1,XCmodel,p1gau,XCmodelgau,Ls2 = GLOBALutils.XC_Final_Fit( vels, xc_av ,\
                                  sigma_res = 4, horder=8, moonv = refvel, moons = moon_sig, moon = False)

                moonmatters = False
                if (know_moon and here_moon):
                    moonmatters = True
                    ismoon = True
                    confused = False
                    p1_m,XCmodel_m,p1gau_m,XCmodelgau_m,Ls2_m = GLOBALutils.XC_Final_Fit( vels, xc_av , sigma_res = 4, horder=8, moonv = refvel, moons = moon_sig, moon = True)
                    moon_flag = 1
                else:
                    confused = False
                    ismoon = False
                    p1_m,XCmodel_m,p1gau_m,XCmodelgau_m,Ls2_m = p1,XCmodel,p1gau,XCmodelgau,Ls2
                    moon_flag = 0

                bspan = GLOBALutils.calc_bss(vels,xc_av)
                SP = bspan[0]

                if (not known_sigma):
                    disp = np.floor(p1gau[2])
                    if (disp < 3.0): 
                        disp = 3.0
                    mask_hw_wide = av_m * disp / (GLOBALutils.Constants.c/1.0e3)
                    ml_v = av_m - mask_hw_wide
                    mh_v = av_m + mask_hw_wide            
                    known_sigma = True
                else:
                    cond = False

            xc_dict = {'vels':vels,'xc_av':xc_av,'XCmodelgau':XCmodelgau,'Ls2':Ls2,'refvel':refvel,\
                       'rvels':rvels,'rxc_av':rxc_av,'rpred':rpred,'rxc_av_orig':rxc_av_orig,\
                       'rvel0_xc':rvel0_xc,'xc_full':xc_full, 'p1':p1, 'sn':sn, 'p1gau':p1gau,\
                       'p1_m':p1_m,'XCmodel_m':XCmodel_m,'p1gau_m':p1gau_m,'Ls2_m':Ls2_m,\
                       'XCmodelgau_m':XCmodelgau_m}

            moon_dict = {'moonmatters':moonmatters,'moon_state':moon_state,'moonsep':moonsep,\
                         'lunation':lunation,'mephem':mephem,'texp':float(h[0].header['EXPTIME'])}

            pkl_xc = dirout + fsim.split('/')[-1][:-8]+obname+'_XC_'+sp_type+'.pkl'
            pickle.dump(xc_dict, open(pkl_xc, 'wb'))

            ccf_pdf = dirout + 'proc/' + fsim.split('/')[-1][:-4] + obname + '_XCs_' + sp_type + '.pdf'

            if not avoid_plot:
                GLOBALutils.plot_CCF(xc_dict,moon_dict,path=ccf_pdf)

            SNR_5130 = np.median(spec[8,103 - difo,900:1101] )
            airmass  = float(h[0].header['AIRMASS'])
            seeing   = -999
            TEXP = float(h[0].header['EXPTIME'])

            if sp_type == 'G2':
                if T_eff < 6000:
                        A = 0.06544
                        B = 0.00146
                        D = 0.24416
                        C = 0.00181
                else:
                        A = 0.09821
                        B = 0.00014
                        D = 0.33491
                        C = 0.00113
            elif  sp_type == 'K5':
                A = 0.05348
                B = 0.00147     
                D = 0.20695
                C = 0.00321
            else:
                A = 0.05348
                B = 0.00147     
                D = 0.20695
                C = 0.00321

            RVerr =  B + ( 1.6 + 0.2 * p1gau[2] ) * A / np.round(SNR_5130)
            depth_fact = 1. + p1gau[0]/(p1gau[2]*np.sqrt(2*np.pi))

            if depth_fact >= 1.:
                RVerr2 = -999.000
            else:
                if sp_type == 'G2':
                    depth_fact = (1 - 0.62) / (1 - depth_fact)
                else:
                    depth_fact = (1 - 0.59) / (1 - depth_fact)
                RVerr2 = RVerr * depth_fact
                if (RVerr2 <= 0.009):
                    RVerr2 = 0.009

            BSerr = D / float(np.round(SNR_5130)) + C

            RV     = np.around(p1gau_m[1],4)  
            BS     = np.around(SP,4)   
            RVerr2 = np.around(RVerr2,4)
            BSerr = np.around(BSerr,4)

            print('\t\t\tRV = '+str(RV)+' +- '+str(RVerr2))
            print('\t\t\tBS = '+str(BS)+' +- '+str(BSerr))

            bjd_out = 2400000.5 + mbjd
            T_eff_err = 100
            logg_err = 0.5
            Z_err = 0.5
            vsini_err = 2
            XC_min = np.abs(np.around(np.min(XCmodel),2))

            SNR_5130 = np.around(SNR_5130)
            SNR_5130_R = np.around(SNR_5130*np.sqrt(2.9))

            disp_epoch = np.around(p1gau_m[2],1)
            hdu = GLOBALutils.update_header(hdu,'RV', RV)
            hdu = GLOBALutils.update_header(hdu,'RV_E', RVerr2)
            hdu = GLOBALutils.update_header(hdu,'BS', BS)
            hdu = GLOBALutils.update_header(hdu,'BS_E', BSerr)
            hdu = GLOBALutils.update_header(hdu,'DISP', disp_epoch)
            hdu = GLOBALutils.update_header(hdu,'SNR', SNR_5130)
            hdu = GLOBALutils.update_header(hdu,'SNR_R', SNR_5130_R)
            hdu = GLOBALutils.update_header(hdu,'INST', 'CAFE')
            hdu = GLOBALutils.update_header(hdu,'RESOL', '70000')
            hdu = GLOBALutils.update_header(hdu,'PIPELINE', 'CERES')
            hdu = GLOBALutils.update_header(hdu,'XC_MIN', XC_min)
            hdu = GLOBALutils.update_header(hdu,'BJD_OUT', bjd_out)

            # write to output
            line_out = "%-15s %18.8f %9.4f %7.4f %9.3f %5.3f   cafe   ceres   70000 %6d %5.2f %5.2f %5.1f %4.2f %5.2f %6.1f %4d %s\n"%\
                      (obname, bjd_out, RV, RVerr2, BS, BSerr, T_eff_epoch, logg_epoch, Z_epoch, vsini_epoch, XC_min, disp_epoch,\
                       TEXP, SNR_5130_R, ccf_pdf)
            f_res.write(line_out)
    
        if (os.access( dirout + fout,os.F_OK)):
            os.remove( dirout + fout)

        hdu.writeto( dirout + fout )
    else:
        print("Reading spectral file from", fout)
        spec = pyfits.getdata( fout )

f_res.close()
